//
// Copyright (C) 1999-2009 WideStudio/MWT Project Team
// Copyright (C) 2006,2007 FUJITSU LIMITED
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#include <WSDdev.h>
#include <WSCdevice.h>
#include <WSCbase.h>
#include <WSDdraw.h>
#include <WSDappDev.h>

WSMFclassInit(WSDdev,WSCroot);

WSCushort WSDdev::_w = 0;
WSCushort WSDdev::_h = 0;
WSDdraw* WSDdev::_draw = NULL;
WSDdraw* WSDdev::_native_draw = NULL;

void WSDdev::getGeom(short* ox,short* oy,WSCushort* ow,WSCushort* oh){
  *ox = _x;
  *oy = _y;
  *ow = _w;
  *oh = _h;
}

WSDdev::WSDdev(){
  _client = NULL;
  _addr_change_handler = 0;
  _handler_data = NULL;
#ifndef NO_SCALE_OFFSET
  _scale_ptr = NULL;
  _x_offset_ptr = NULL;
  _y_offset_ptr = NULL;
#endif
  _absolute_draw = False;
  _no_child = False;
  _parent_dev = NULL;
  _event_parent_dev = NULL;
  _registered = False;
  _event_mask = 0;
  _w_x = 0;
  _w_y = 0;
  _x = 0;
  _y = 0;
  _deleted = 0;
  _visible = 0;
  _operation = WS_OPERATION_SRC_COPY;
  _pixmap_style = WS_DIRECT_WINDOW;
  _expose_sequence = False;
  _draw_pixmap_only = False;
  _draw_pixmap_only_bk = False;
  _draw_pixmap_only_ch = False;
  setInitialized(0);
}

WSDdev::~WSDdev(){
  detachClient();
  setInitialized(False);
}

WSCbool WSDdev::getNoChildDev(){
  return _no_child;
}

void WSDdev::setEventMaskBit(long mask){
  _event_mask = mask;
}

long WSDdev::getEventMaskBit(){
  return _event_mask;
}

long WSDdev::setAbsoluteAddrChangeHandler(void(*hd)(WSDdev*,void*),void* data){
  _handler_data = data;
  _addr_change_handler = hd;
  return WS_NO_ERR;
}

WSCbool WSDdev::getEventRegistered(){
  return _registered;
}

void WSDdev::setEventRegistered(WSCbool fl){
  if (fl != False){
    fl = True;
  }else{
    setEventMaskBit(0);
  }
  _registered = fl;
}

WSCbool WSDdev::getReady(){
  return False;
}
#ifndef NO_SCALE_OFFSET
void WSDdev::setScalePtr(double* ptr){
  _scale_ptr = ptr;
}

void WSDdev::setXOffsetPtr(short* ptr){
  _x_offset_ptr = ptr;
}

void WSDdev::setYOffsetPtr(short* ptr){
  _y_offset_ptr = ptr;
}
#endif
WSCbool WSDdev::getScaled(){
#ifndef NO_SCALE_OFFSET
  if (_scale_ptr == NULL){
    return False;
  }
  if (*_scale_ptr == 1.0){
    return False;
  }
  return True;
#else
  return False;
#endif
}
double WSDdev::getScale(){
#ifndef NO_SCALE_OFFSET
  if (_scale_ptr == NULL){
    return (double)1;
  }
  if (*_scale_ptr < 0){
    return (double)0;
  }
  return *_scale_ptr; 
#else
  return (double)1;
#endif
}

void WSDdev::getParentPtrOffset(short* xoff,short* yoff){
  WSDdev* pdev = getParentDev();
  if (pdev != NULL &&
      pdev->getIsNonWindow() != False){
    if (pdev->_x_offset_ptr != NULL){
      *xoff += *pdev->_x_offset_ptr;
    }
    if (pdev->_y_offset_ptr != NULL){
      *yoff += *pdev->_y_offset_ptr;
    }
    pdev->getParentPtrOffset(xoff,yoff);
  }
}
void WSDdev::getOffset(short* xoff,short* yoff){
  short diffx = 0;
  short diffy = 0;
  getNonWindowOffset(&diffx,&diffy);
  getParentPtrOffset(&diffx,&diffy);
#ifndef NO_SCALE_OFFSET
  if (_x_offset_ptr == NULL){
    *xoff = diffx;
  }else{
    *xoff = *_x_offset_ptr + diffx; 
  }
  if (_y_offset_ptr == NULL){
    *yoff = diffy;
  }else{
    *yoff = *_y_offset_ptr + diffy; 
  }
#else
  *xoff = diffx;
  *yoff = diffy;
#endif
}

short WSDdev::getXOffset(){
  short diffx = 0;
  short diffy = 0;
  getNonWindowOffset(&diffx,&diffy);
  getParentPtrOffset(&diffx,&diffy);
#ifndef NO_SCALE_OFFSET
  if (_x_offset_ptr == NULL){
    return diffx;
  }
  return *_x_offset_ptr + diffx; 
#else
  return diffx;
#endif
}

short WSDdev::getYOffset(){
  short diffx = 0;
  short diffy = 0;
  getNonWindowOffset(&diffx,&diffy);
  getParentPtrOffset(&diffx,&diffy);
#ifndef NO_SCALE_OFFSET
  if (_y_offset_ptr == NULL){
    return diffy;
  }
  return *_y_offset_ptr + diffy; 
#else
  return diffy;
#endif
}

long WSDdev::setValue(long kind ,void* data){
#ifndef NO_SCALE_OFFSET
  switch(kind){
    case WSDEV_SCALE_OFFSET:
             setScalePtr((double*)data);
             break;
    case WSDEV_X_OFFSET:
             setXOffsetPtr((short*)data);
             break;
    case WSDEV_Y_OFFSET:
             setYOffsetPtr((short*)data);
             break;
  }
#endif
  return WS_NO_ERR;
}

long WSDdev::getValue(long /*kind*/,void* /*value*/){
  return WS_NO_ERR;
}

long WSDdev::getAvailableEventBit(){
  return 0;
}

WSCbool WSDdev::isAvailableEvent(long /*ev*/){
  return False;
}

long WSDdev::setDisableEvent(long /*ev*/){
  return WS_NO_ERR;
}

long WSDdev::setEventOrder(WSDdev*,char /*top_or_bottom*/){
  return WS_NO_ERR;
}

long WSDdev::setEvent(WSDdev*,WSCbool /*on_off*/){
  return WS_NO_ERR;
}

long WSDdev::execEvent(long /*ev*/,void* /*data*/){
  return WS_NO_ERR;
}

WSCbool WSDdev::isExposed(short x,short y,WSCushort w,WSCushort h,WSCbool scaling){
  if (scaling != False){
    WSCbool scaled = getScaled();
    short xoff;
    short yoff;
    getOffset(&xoff,&yoff);
    if ( scaled != False){
      double scale = getScale();
      x = (short)((x + xoff)* scale);
      y = (short)((y + yoff)* scale);
      w = (WSCushort)(w * scale);
      h = (WSCushort)(h * scale);
    }else{
      x = x + xoff;
      y = y + yoff;
    }
  }

  short ex,ey;
  WSCushort ew,eh;
  getExposedArea(&ex,&ey,&ew,&eh,False);
  if (x + w < ex || y + h < ey || x > ex + ew || y > ey + eh ){
    return False;
  }
  return True;
}

long WSDdev::informEventToClient(long ev_kind,void* data){
  if (_client == NULL){
    return WS_ERR;
  }
  if (_client->getVisible() == False){
    return WS_ERR;
  }
  if (ev_kind == WSEV_EXPOSE){  
    return _client->execEventProc(ev_kind,data);
  }
#ifndef NO_MODAL
  WSCbase* mcl = WSGIappDev()->getModalClient();
  if (mcl != NULL && mcl != _client && ev_kind != WSEV_RESIZE){
//WSMFtrace("WSDdev::informEventToClient modal\n");
    WSCbool val = False;
    WSDdev* d = this;
    while(d != NULL){
      d->getValue(WSDEV_GRAB_POINTER,&val);
      if (val != False){
        break;
      }
      d = d->getParentDev();
    }
    if (val == False){
      void(*proc)(WSDdev*,long,void*) = (void(*)(WSDdev*,long,void*))WSGIappDev()->getModalProc();
      WSCbase* cl = _client->getParent();
      if (cl == NULL){
        if (proc != NULL){
          proc(this,ev_kind,data);
          return WS_NO_ERR;
        }
        return WS_ERR;
      }
      long fl = False;
      while(1){
        if (cl == mcl){
          fl = True;
          break;
        }
        cl = cl->getParent();
        if (cl == NULL){
          break;
        }
      }
      if (fl == False){
        if (proc != NULL){
          proc(this,ev_kind,data);
          return WS_NO_ERR;
        }
        return WS_ERR;
      }
    }
  }
  WSCbase* pmcl = WSGIappDev()->getPopupModalClient();
  if (pmcl != NULL && pmcl != _client){
    return WS_ERR;
  }
#endif
  if (ev_kind == WSEV_MOUSE_IN || ev_kind == WSEV_MOUSE_MOVE ||
      ev_kind == WSEV_MOUSE_DOUBLE_CLICK ||
      ev_kind == WSEV_MOUSE_PRESS || ev_kind == WSEV_MOUSE_RELEASE){
     WSCpoint* src = (WSCpoint*)data;
     WSDdev* pdev = getParentDev();
     WSDdev* _this = this;
     while(1){
       if (_this->getIsNonWindow() == False){
         break;
       }
       if (pdev != NULL &&
         pdev->getIsNonWindow() != False &&
         pdev->clipChildren() != False &&
         _this->clipping() != False){
         long cx = pdev->_w_x;
         long cy = pdev->_w_y;
         WSCbase* pclient = pdev->getAttachedClient();
         if (pclient != NULL){
           WSCulong cw = pclient->getProperty(WSNwidth);
           WSCulong ch = pclient->getProperty(WSNheight);
           double pscale = pdev->getScale();
           short xoff;
           short yoff;
           pdev->getOffset(&xoff,&yoff);
           cx = (short)((cx + xoff)*pscale);
           cy = (short)((cy + yoff)*pscale);
           cw = (WSCushort)(cw * pscale);
           ch = (WSCushort)(ch * pscale);

           if (src->x < cx || cx+cw < src->x ||
             src->y < cy || cy+ch < src->y){
             if ((_client->getOutSideMousePress() == True &&
                  (ev_kind == WSEV_MOUSE_PRESS ||
                   ev_kind == WSEV_MOUSE_DOUBLE_CLICK)) ||
                 (_client->getOutSideMouseMove() == True &&
                  ev_kind == WSEV_MOUSE_MOVE) ||
                 (_client->getOutSideMouseRelease() == True &&
                  ev_kind == WSEV_MOUSE_RELEASE)){
               //OK..
             }else{
               return WS_ERR;
             }
           }
         }
       }
       if (pdev == NULL){
         break;
       }
       _this = pdev;
       pdev = pdev->getParentDev();
       if (pdev == NULL || pdev->getIsNonWindow() == False){
         break;
       }
     }

     WSCpoint point;
     WSCbool scaled = getScaled();
     short xoff;
     short yoff;
     getOffset(&xoff,&yoff);
     if (scaled != False){
       double scale = getScale();
       point.setPoint((short)(src->x / scale) - xoff,
                      (short)(src->y / scale) - yoff); 
     }else{
       point.setPoint(src->x - xoff,src->y - yoff); 
     }

#ifndef WS_EMBED
     if (_client->getSensitive() != False
         || _client->getPropertyEditMode() != False){
       return _client->execEventProc(ev_kind,&point);
     }
#else
     if (_client->getSensitive() != False){
       return _client->execEventProc(ev_kind,&point);
     }
#endif
     return WS_ERR;
  }
  if ( ev_kind == WSEV_MOUSE_OUT){
#ifndef WS_EMBED
     if (_client->getSensitive() != False
         || _client->getPropertyEditMode() != False){
       return _client->execEventProc(ev_kind,data);
     }
#else
     if (_client->getSensitive() != False){
       return _client->execEventProc(ev_kind,data);
     }
#endif
     return WS_ERR;
  }
  return  _client->execEventProc(ev_kind,data);
}

void WSDdev::setInitialized(WSCbool fl){
   _initialized = fl;
}

WSCbool WSDdev::getInitialized(){
   return  _initialized;
}

long WSDdev::attachClient(WSCbase* client){
   if (_client != NULL){
     detachClient();
   }
#ifndef NO_SCALE_OFFSET
   _x_offset_ptr = NULL;
   _y_offset_ptr = NULL;
   _scale_ptr = NULL;
#endif
   _client = client;

   if (getInitialized() == False){
     initialize();
     setInitialized(True);
   }
   return WS_NO_ERR;
}

long WSDdev::initialize(){
  setInitialized(True);
  return WS_NO_ERR;
}

long WSDdev::cancelEvent(WSDdev* dev){
  if (dev == this){
    setEventRegistered(False);
  }
  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){
    pdev->cancelEvent(this);
  }
  return WS_NO_ERR;
}

long WSDdev::detachClient(){
   if (_client == NULL){
     return WS_NO_ERR;
   }
   setVisible(False);
   cancelEvent(this);
   _addr_change_handler = NULL;
   _handler_data = NULL;

   _client = NULL;
   _parent_dev = NULL;
   _event_parent_dev = NULL;

   return WS_NO_ERR;
}

WSCbool WSDdev::getVisible(){
   return _visible;
}
long WSDdev::setVisible(WSCbool fl){
  _visible = fl;
  if (fl != False && getInitialized() == False){
    initialize();
    setInitialized(True);
  }
  return WS_NO_ERR;
}


long  WSDdev::setDraw(WSDdraw*  draw){
  _draw = draw;
  if (_native_draw == NULL){
    _native_draw = draw;
  }
  return WS_NO_ERR;
}
WSDdraw* WSDdev::getDraw(){
  return _draw;
}
long  WSDdev::setNativeDraw(WSDdraw*  draw){
  _native_draw = draw;
  return WS_NO_ERR;
}
WSDdraw* WSDdev::getNativeDraw(){
  return _native_draw;
}

long WSDdev::internalBeginDraw(short x,short y,WSCushort w,WSCushort h,WSCbool absolute,WSCbool scaling){
  return beginDraw(x,y,w,h,absolute,scaling);
}
long WSDdev::beginDraw(short x,short y,WSCushort w,WSCushort h,WSCbool absolute,WSCbool scaling){
  _absolute_draw = absolute;
  if (scaling != False){
    short xoff;
    short yoff;
    getOffset(&xoff,&yoff);
    if (getScaled() == False){
      x += xoff;
      y += yoff;
    }else{
      double scale = getScale();
      x = (short)((x + xoff)*scale);
      y = (short)((y + yoff)*scale);
      w = (WSCushort)(w * scale);
      h = (WSCushort)(h * scale);
    }
  }

  WSDdev* pdev = getParentDev();
  if ( pdev != NULL ){
      pdev->beginDraw( x, y, w, h,absolute,False);
  } 

  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }

  _x = x; 
  _y = y; 
  _w = w; 
  _h = h; 

  short ex,ey;
  WSCushort ew,eh;
  long ret = getExposedArea(&ex,&ey,&ew,&eh,False);
  if (ret != WS_NO_ERR){
    return WS_ERR;
  }

  if (getReady() == False){
    return WS_ERR;
  }
  draw->beginDraw(this,getDeviceResource(),getWindowResource(),getContextResource());

  short ex2 = ex + ew < x + w ? ex + ew : x + w; 
  short ey2 = ey + eh < y + h ? ey + eh : y + h; 
  ex = ex > x ? ex : x; 
  ey = ey > y ? ey : y; 
  if (ex2 > ex){
    ew = ex2 - ex;
  }else{
    ew = 0;
  }
  if (ey2 > ey){
    eh = ey2 - ey;
  }else{
    eh = 0;
  }

  if (absolute == False){
    draw->setRegion(ex,ey,ew,eh);
  }else{
    draw->setRegion(x,y,w,h);
  }

  return WS_NO_ERR;
}
long WSDdev::internalEndDraw(){
  return endDraw();
}
long WSDdev::endDraw(){
  _absolute_draw = False;
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  draw->endDraw();
  WSDdev* pdev = getParentDev();
  if ( pdev != NULL ){
    pdev->endDraw();
  } 
  return WS_NO_ERR;
}

long WSDdev::setLineWidth(short w){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  if (w < 0){
    w = 0;
  }
  draw->setLineWidth(w);
  return WS_NO_ERR;
}

long WSDdev::setLineDashType(char type){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  draw->setLineDashType(type);
  return WS_NO_ERR;
}
#ifndef NO_HATCHING
long WSDdev::setHatchPattern(char type){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  draw->setHatchPattern(type);
  return WS_NO_ERR;
}
#endif
long WSDdev::setForeColor(short no){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  draw->setForeColor(no);
  return WS_NO_ERR;
}

long WSDdev::setBackColor(short no){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  draw->setBackColor(no);
  return WS_NO_ERR;
}
long WSDdev::setForeColor(WSDcolor* col){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  draw->setForeColor(col);
  return WS_NO_ERR;
}

long WSDdev::setBackColor(WSDcolor* col){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  draw->setBackColor(col);
  return WS_NO_ERR;
}

long WSDdev::drawArc(short x,short y,WSCushort w,WSCushort h,long a1,long a2){
  if (w < 1){
    return WS_NO_ERR;
  }
  if (h < 1){
    return WS_NO_ERR;
  }
  if (a2 == 0){
    return WS_NO_ERR;
  }
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawArc((short)(x * scale) +_x, 
                  (short)(y * scale) +_y,
                  (WSCushort)(w * scale), 
                  (WSCushort)(h * scale), a1, a2);
  }else{
    draw->drawArc(x +_x, y +_y, w , h , a1, a2);
  }
  return WS_NO_ERR;
}

long WSDdev::drawFillArc(short x,short y,WSCushort w,WSCushort h,long a1,long a2,char kind){
  if (w < 1){
    return WS_NO_ERR;
  }
  if (h < 1){
    return WS_NO_ERR;
  }
  if (a2 == 0){
    return WS_NO_ERR;
  }
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }

  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawFillArc((short)(x * scale) +_x, 
                      (short)(y * scale) +_y,
                      (WSCushort)(w * scale), 
                      (WSCushort)(h * scale), a1, a2,kind);
  }else{
    draw->drawFillArc(x +_x, y +_y, w , h , a1, a2,kind);
  }
  return WS_NO_ERR;
}

long WSDdev::drawLines(WSCpoint* points,long num){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }

  WSCpoint  p1[16];
  WSCpoint* pt;
  if (num > 16){
    pt = new WSCpoint[num];
  }else{
    pt = p1;
  }

  long i;
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    for(i = 0; i < num; i++){
      pt[i].x = (short)(points[i].x * scale) +_x;
      pt[i].y = (short)(points[i].y * scale) +_y;
    };
  }else{
    for(i = 0; i < num; i++){
      pt[i].x = points[i].x +_x;
      pt[i].y = points[i].y +_y;
    };
  }
  draw->drawLines(pt,num);
  if (num > 16){
    delete pt;
  }
  return WS_NO_ERR;
}

long WSDdev::drawLine(short x1,short y1,short x2,short y2){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }

  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawLine((short)(x1 * scale) +_x, (short)(y1 * scale) +_y,
                   (WSCushort)(x2 * scale) +_x,
                   (WSCushort)(y2 * scale) +_y);
  }else{
    draw->drawLine(x1 +_x, y1 +_y, x2 +_x, y2 +_y);
  }
  return WS_NO_ERR;
}

long WSDdev::drawRect(short x,short y,WSCushort w,WSCushort h){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    WSCushort vw = (WSCushort)(w * scale);
    WSCushort vh = (WSCushort)(h * scale);
    if (vw < 1){
      vw = 1;
    }
    if (vh < 1){
      vh = 1;
    }
    draw->drawRect((short)(x * scale) +_x, (short)(y * scale) +_y,vw,vh);
  }else{
    if (w < 1){
      w = 1;
    }
    if (h < 1){
      h = 1;
    }
    draw->drawRect(x +_x, y +_y,w,h);
  }
  return WS_NO_ERR;
}
long WSDdev::drawFillRect(short x,short y,WSCushort w,WSCushort h){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    WSCushort vw = (WSCushort)(w * scale);
    WSCushort vh = (WSCushort)(h * scale);
    if (vw < 1){
      vw = 1;
    }
    if (vh < 1){
      vh = 1;
    }
    draw->drawFillRect((short)(x * scale) +_x, (short)(y * scale) +_y,
                       (WSCushort)(w * scale),
                       (WSCushort)(h * scale));
  }else{
    if (w < 1){
      w = 1;
    }
    if (h < 1){
      h = 1;
    }
    draw->drawFillRect(x+_x,y+_y,w,h);
  }
  return WS_NO_ERR;
}
long WSDdev::drawGradation(long type,short col1,short col2,short col3,
                   short x,short y,WSCushort w,WSCushort h,WSCuchar gradm){
#ifndef NO_GRADATION
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawGradation(type,col1,col2,col3,
                       (short)(x * scale) +_x, (short)(y * scale) +_y,
                       (WSCushort)(w * scale),
                       (WSCushort)(h * scale),gradm);
  }else{
    draw->drawGradation(type,col1,col2,col3,x+_x,y+_y,w,h,gradm);
  }
#endif
  return WS_NO_ERR;
}



long WSDdev::exposeArea(long x,long y,WSCulong w,WSCulong h,WSCbool scaling){
  return WS_NO_ERR;
}
long WSDdev::clearArea(long x,long y,WSCulong w,WSCulong h,WSCbool create_event,WSCbool scaling){
  return WS_NO_ERR;
}
long WSDdev::erase(long x,long y,WSCulong w,WSCulong h,WSCbool create_event,WSCbool scaling){
  return WS_NO_ERR;
}
long WSDdev::copyArea(long x,long y,WSCulong w,WSCulong h,long dx,long dy){
#if 0
  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){

    double scale = getScale();
    if (scaling != False && scale != 1){
        pdev->clearArea((short)(x * scale) +_x,(short)(y * scale) +_y,
                      (WSCushort)(w * scale), (WSCushort)(h * scale),
                      (short)(dx * scale) +_x,(short)(dy * scale) +_y);
    }else{
      pdev->clearArea(x + _x,y + _y, w, h, dx + _x, dy + _y);
    }
  }
#endif
  return WS_NO_ERR;
}

void WSDdev::setExposedArea(short,short,WSCushort,WSCushort){
}

long WSDdev::setEnableEventBit(long /*ev_bit*/){
  setEventRegistered(True);
  return WS_NO_ERR;
}

long WSDdev::setEnableEvent(long /*ev_kind*/){
  setEventRegistered(True);
  return WS_NO_ERR;
}

long WSDdev::drawRects(WSCrect* rects,long num){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCrect  r1[16];
  WSCrect* rt;
  if (num > 16){
    rt = new WSCrect[num];
  }else{
    rt = r1;
  }

  long i;
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    for(i = 0; i < num; i++){
      rt[i].x = (short)(rects[i].x * scale) +_x;
      rt[i].y = (short)(rects[i].y * scale) +_y;
      rt[i].width = (WSCushort)(rects[i].width * scale);
      rt[i].height = (WSCushort)(rects[i].height * scale);
      if (rt[i].width < 1){
        rt[i].width = 1;
      }
      if (rt[i].height < 1){
        rt[i].height = 1;
      }
    };
  }else{
    for(i = 0; i < num; i++){
      rt[i].x = rects[i].x +_x;
      rt[i].y = rects[i].y +_y;
      rt[i].width = rects[i].width;
      rt[i].height = rects[i].height;
      if (rt[i].width < 1){
        rt[i].width = 1;
      }
      if (rt[i].height < 1){
        rt[i].height = 1;
      }
    };
  }
  draw->drawRects(rt,num);
  if (num > 16){
    delete rt;
  }
  return WS_NO_ERR;
}

long WSDdev::drawFillRects(WSCrect* rects,long num){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCrect  r1[16];
  WSCrect* rt;
  if (num > 16){
    rt = new WSCrect[num];
  }else{
    rt = r1;
  }
  long i;
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    for(i = 0; i < num; i++){
      rt[i].x = (short)(rects[i].x * scale) +_x;
      rt[i].y = (short)(rects[i].y * scale) +_y;
      rt[i].width = (WSCushort)(rects[i].width * scale);
      rt[i].height = (WSCushort)(rects[i].height * scale);
      if (rt[i].width < 1){
        rt[i].width = 1;
      }
      if (rt[i].height < 1){
        rt[i].height = 1;
      }
    };
  }else{
    for(i = 0; i < num; i++){
      rt[i].x = rects[i].x +_x;
      rt[i].y = rects[i].y +_y;
      rt[i].width = rects[i].width;
      rt[i].height = rects[i].height;
      if (rt[i].width < 1){
        rt[i].width = 1;
      }
      if (rt[i].height < 1){
        rt[i].height = 1;
      }
    };
  }
  draw->drawFillRects(rt,num);
  if (num > 16){
    delete rt;
  }
  return WS_NO_ERR;
}

long WSDdev::drawPoly(WSCpoint* points,long num){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCpoint  p1[16];
  WSCpoint* pt;
  if (num > 16){
    pt = new WSCpoint[num];
  }else{
    pt = p1;
  }
  long i;
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    for(i = 0; i < num; i++){
      pt[i].x = (short)(points[i].x * scale) +_x;
      pt[i].y = (short)(points[i].y * scale) +_y;
    };
  }else{
    for(i = 0; i < num; i++){
      pt[i].x = points[i].x +_x;
      pt[i].y = points[i].y +_y;
    };
  }
  draw->drawPoly(pt,num);
  if (num > 16){
    delete pt;
  }
  return WS_NO_ERR;
}

long WSDdev::drawFillPoly(WSCpoint* points,long num){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCpoint  p1[16];
  WSCpoint* pt;
  if (num > 16){
    pt = new WSCpoint[num];
  }else{
    pt = p1;
  }

  long i;
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    for(i = 0; i < num; i++){
      pt[i].x = (short)(points[i].x * scale) +_x;
      pt[i].y = (short)(points[i].y * scale) +_y;
    }
  }else{
    for(i = 0; i < num; i++){
      pt[i].x = points[i].x +_x;
      pt[i].y = points[i].y +_y;
    }
  }
  draw->drawFillPoly(pt,num);
  if (num > 16){
    delete pt;
  }
  return WS_NO_ERR;
}
#ifndef NO_IMAGE
long WSDdev::drawImage(short x,short y,WSCushort w,WSCushort h,WSDimage* image,char align){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawImage( (short)(x * scale) +_x,
                     (short)(y * scale) +_y,
                     (WSCushort)(w * scale),
                     (WSCushort)(h * scale),image,align);
  }else{
    draw->drawImage(x+_x,y+_y,w,h,image,align);
  }
  return WS_NO_ERR;
}
long WSDdev::drawImage(short x,short y,short sx,short sy,WSCushort w,WSCushort h,WSDimage* image){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawImage( (short)(x * scale) +_x,
                     (short)(y * scale) +_y,
                     sx,sy,
                     (WSCushort)(w * scale),
                     (WSCushort)(h * scale),image);
  }else{
    draw->drawImage(x+_x,y+_y,sx,sy,w,h,image);
  }
  return WS_NO_ERR;
}
long WSDdev::drawStretchedImage(short x,short y,WSCushort w,WSCushort h,WSDimage* image){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawStretchedImage( (short)(x * scale) +_x,
                     (short)(y * scale) +_y,
                     (WSCushort)(w * scale),
                     (WSCushort)(h * scale),image);
  }else{
    draw->drawStretchedImage(x+_x,y+_y,w,h,image);
  }
  return WS_NO_ERR;
}
#endif
long WSDdev::drawString(long x,long y,WSCulong w,WSCulong h,WSCstring* str,char font,char align,long cur,WSCbool inter_cur,long spos1,long spos2){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawString( (long)(x * scale) +_x, 
                      (long)(y * scale) +_y,
                      (WSCulong)(w * scale), 
                      (WSCulong)(h * scale),
                      str,font,align,cur,inter_cur,
                      spos1,spos2);
  }else{
    draw->drawString(x+_x,y+_y,w,h,str,font,align,cur,inter_cur,spos1,spos2);
  }
  return WS_NO_ERR;
}

long WSDdev::drawString(long x,long y,WSCulong w,WSCulong h,char font,char align,char* str,long encoding){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCstring tmp;
  tmp.setStaticString(str,encoding);
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawString( (long)(x * scale) +_x, 
                      (long)(y * scale) +_y,
                      (WSCulong)(w * scale), 
                      (WSCulong)(h * scale),
                      &tmp,font,align,-1,False,0,0);
  }else{
    draw->drawString(x+_x,y+_y,w,h,&tmp,font,align,-1,False,0,0);
  }
  return WS_NO_ERR;
}

long WSDdev::drawUString(long x,long y,WSCulong w,WSCulong h,WSCushort* str,char font,char align,long cur,WSCbool inter_cur,long spos1,long spos2){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawUString( (long)(x * scale) +_x, 
                      (long)(y * scale) +_y,
                      (WSCulong)(w * scale), 
                      (WSCulong)(h * scale),
                      str,font,align,cur,inter_cur,
                      spos1,spos2);
  }else{
    draw->drawUString(x+_x,y+_y,w,h,str,font,align,cur,inter_cur,spos1,spos2);
  }
  return WS_NO_ERR;
}

long WSDdev::drawFillString(long x,long y,WSCulong w,WSCulong h,WSCstring* str,char font,char align,long cur,WSCbool inter_cur){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawFillString( (long)(x * scale) +_x, 
                          (long)(y * scale) +_y,
                          (WSCulong)(w * scale), 
                          (WSCulong)(h * scale),
                          str,font,align,cur,inter_cur);
  }else{
    draw->drawFillString(x+_x,y+_y,w,h,str,font,align,cur,inter_cur);
  }
  return WS_NO_ERR;
}

long WSDdev::drawFillString(long x,long y,WSCulong w,WSCulong h,char font,char align,char* str,long encoding){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCstring tmp;
  tmp.setString(str,encoding);
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawFillString( (long)(x * scale) +_x, 
                          (long)(y * scale) +_y,
                          (WSCulong)(w * scale), 
                          (WSCulong)(h * scale),
                          &tmp,font,align,-1,False);
  }else{
    draw->drawFillString(x+_x,y+_y,w,h,&tmp,font,align,-1,False);
  }
  return WS_NO_ERR;
}

long WSDdev::drawFillUString(long x,long y,WSCulong w,WSCulong h,WSCushort* str,char font,char align,long cur,WSCbool inter_cur){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  WSCbool scaled = getScaled();
  if (scaled != False){
    double scale = getScale();
    draw->drawFillUString( (long)(x * scale) +_x, 
                          (long)(y * scale) +_y,
                          (WSCulong)(w * scale), 
                          (WSCulong)(h * scale),
                          str,font,align,cur,inter_cur);
  }else{
    draw->drawFillUString(x+_x,y+_y,w,h,str,font,align,cur,inter_cur);
  }
  return WS_NO_ERR;
}

long WSDdev::setRegion(short x,short y,WSCushort w, WSCushort h){
  WSDdraw* draw = getDraw();

  short ex,ey;
  WSCushort ew,eh;
  long ret = getExposedArea(&ex,&ey,&ew,&eh,False);
  if (ret != WS_NO_ERR){
    return WS_ERR;
  }

  WSCbool scaled = getScaled();

  if (scaled != False){
    double scale = getScale();
    x=(short)(x * scale) +_x;
    y=(short)(y * scale) +_y;
    w=(WSCushort)(w * scale);
    h=(WSCushort)(h * scale);
  }else{
    x+=_x;
    y+=_y;
  }

  long ox,oy;
  WSCulong ow,oh;
  WSGFandArea(x,y,w,h,ex,ey,ew,eh,&ox,&oy,&ow,&oh);

  if (_absolute_draw == False){
    x = ox;
    y = oy;
    w = ow;
    h = oh;
  }

  draw->setRegion(x,y,w,h);

  return WS_NO_ERR;
}

long WSDdev::getExposedArea(short* x,short* y, WSCushort* w, WSCushort* h,WSCbool scaling){
  WSDdev* dev = getEventParentDev();
  if (dev != NULL){
      dev->getExposedArea(x,y,w,h,False);
      if (scaling != False){
        WSCbool scaled = getScaled();
        short xoff;
        short yoff;
        getOffset(&xoff,&yoff);
        if (scaled != False){
          double scale = getScale();
          *x = (short)(*x / scale) - xoff;
          *y = (short)(*y / scale) - yoff;
          *w = (WSCushort)(*w / scale);
          *h = (WSCushort)(*h / scale);
        }else{
          *x = *x  - xoff;
          *y = *y  - yoff;
          *w = *w;
          *h = *h;
        }
      }
      return WS_NO_ERR;
  }else{
      return WS_ERR;
  }
}

void WSDdev::setParentDev(WSDdev* dev){
  _parent_dev = dev;
}
void WSDdev::setEventParentDev(WSDdev* dev){
  _event_parent_dev = dev;
}
WSDdev* WSDdev::getVisibleParentDev(){
  return _parent_dev;
}

WSDdev* WSDdev::getEventParentDev(){
  if (_parent_dev == NULL){
    getParentDev();
  }
  if (_event_parent_dev == NULL){
    _event_parent_dev = _parent_dev;
    if (_event_parent_dev != NULL){
      while(1){
        if (_event_parent_dev != NULL &&
            _event_parent_dev->getIsNonWindow() != False){
          _event_parent_dev = _event_parent_dev->getParentDev();
        }else{
          break;
        }
      }
    }
  }
  return _event_parent_dev;
}

WSDdev* WSDdev::getParentDev(){
  if (_parent_dev != NULL){
    return _parent_dev;
  }

  if (_client == NULL){
    return NULL;
  }
  _parent_dev = _client->getParentDev();
//  if (_parent_dev == this){
//    assert(0);
//  }
  return _parent_dev;
}

WSCbase* WSDdev::getAttachedClient(){
  return _client;
}

void WSDdev::getDispAddr(short* ox,short* oy){
   *ox = 0;
   *oy = 0;
   return;
}

void WSDdev::getAbsoluteAddr(short x,short y,short* ox,short* oy){
  short xoff;
  short yoff;
  getOffset(&xoff,&yoff);
  *ox = (short)((x + xoff) * getScale());
  *oy = (short)((y + yoff) * getScale());
}

WSCbool WSDdev::getAttached(){
  if (getAttachedClient() == NULL){
    return False;
  }
  return True;
}

#if 0
void WSDdev::setDependDev(WSDdev*){
}

WSDdev* WSDdev::getDependDev(){
  return NULL;
}

WSCbool WSDdev::isIndependent(){
  return True;
}
#endif

void WSDdev::deleteInstance(){
  if (_deleted == 0){
    detachClient();
    setInitialized(False);
    _deleted = 1;
    WSGIappDevice()->addDeleteInstance(this);
  }
}
void WSDdev::getWH(WSCushort* w,WSCushort* h){
  *w = _w;
  *h = _h;
}
long WSDdev::setOperation(WSCuchar op){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  _operation = op;
  draw->setOperation(op);
  return WS_NO_ERR;
}
WSCuchar WSDdev::getOperation(){
  return _operation;
}
void WSDdev::setExposeSequence(WSCbool fl){
  _expose_sequence = fl;
}
WSCbool WSDdev::getExposeSequence(){
  return _expose_sequence;
}   
char WSDdev::getPixmapStyle(){
  return _pixmap_style;
}

void WSDdev::setUnderLineChar(long ustart,long ulen){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return;
  }
  draw->setUnderLineChar(ustart,ulen);
}
void WSDdev::setDrawPixmapOnly(WSCbool fl){
  _draw_pixmap_only = fl;
}
WSCbool WSDdev::getDrawPixmapOnly(){
  return _draw_pixmap_only;
}
void WSDdev::onExpose(WSCrect*){
}
WSCbool WSDdev::getIsNonWindow(){
  return False;
}
void WSDdev::getNonWindowOffset(short* ox,short* oy){
  if (getIsNonWindow() == False){
    *ox = 0;
    *oy = 0;
    return;
  }
  WSDdev* parent_dev = getParentDev();

  if (parent_dev == NULL || parent_dev->getIsNonWindow() == False){
    *ox = 0;
    *oy = 0;
    return;
  }

  short diffx = 0;
  short diffy = 0;
  while(1){
    if (parent_dev->getIsNonWindow() != False){
      diffx += parent_dev->_w_x;
      diffy += parent_dev->_w_y;
      parent_dev = parent_dev->getParentDev();
      if (parent_dev == NULL){
        break;
      }
    }else{
      break;
    }
  }
  *ox = diffx;
  *oy = diffy;
  return;
}
WSCbool WSDdev::clipping(){
  return True;
}
WSCbool WSDdev::clipChildren(){
  return True;
}
void (*_event_dispatched_handler)(void*) = NULL;
void* _event_dispatched_handler_data = NULL;
void* WSDdev::getEventDispatchedHandler(){
  return (void*)_event_dispatched_handler;
};
void* WSDdev::getEventDispatchedHandlerParam(){
  return _event_dispatched_handler_data;
}
long WSDdev::setEventDispatchedHandler(void(*hd)(void*),void* ptr){
  _event_dispatched_handler = hd;
  _event_dispatched_handler_data = ptr;
  return WS_NO_ERR;
}
long WSDdev::execEventDispatchedHandler(){
  if (_event_dispatched_handler == NULL){
    return WS_ERR;
  }
  _event_dispatched_handler(_event_dispatched_handler_data);
  _event_dispatched_handler = NULL;
  _event_dispatched_handler_data = NULL;
  return WS_NO_ERR;
}
