//
// Copyright (C) 1999-2008 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#include <WSDmouse.h>
#include <WSCstring.h>
#include <WSDappDev.h>
#include <WSCbase.h>

WSMFclassInit(WSDmouse,WSCroot);

static WSDmouse* (*_chandler)() = NULL;
void WSDmouse::setCreateInstanceHandler( WSDmouse* (*hd)()){
  _chandler = hd;
}
static WSDmouse* _the_app_mouse = NULL;
WSDmouse* WSGIappMouse(){
  if (_the_app_mouse == NULL){
    if (_chandler == NULL){
      return NULL;
WSMFtrace("WSGImouse() ERROR. cannot create WSDmouse class instance.\n");
    }
    _the_app_mouse = _chandler();
  }
  return _the_app_mouse;
}

WSCmouseData::WSCmouseData(){
  id = 0;
  data = 0;
  name = NULL;
};
void WSCmouseData::setName(char* n){
  if (name != NULL){
    delete[] name;
  }
  name = WSGFstrdup(n);
}
WSCmouseData::~WSCmouseData(){
  if (name != NULL){
    delete[] name;
  }
}


WSDmouse::WSDmouse(){
  _status = 0;
  _target_btn = 0 ;
  _global_proc = NULL;
  _special_global_proc = NULL;
  _global_mouse_id = -1;
  _global_mouse_sensitive = True;
}
void WSDmouse::initialize(){}

void WSDmouse::setStatus(WSCulong status){
  _status = status;
}
WSCulong WSDmouse::getStatus(){
  return _status;
}
void WSDmouse::setTargetBtn(WSCulong no){
  _target_btn = no;
}
WSCulong WSDmouse::getTargetBtn(){
  return _target_btn;
}
void WSDmouse::delMouseCursor(long pos){
  if (pos < 0){
    return;
  }
  long num = _mlist.getNum();
  if (pos < num){
    WSCmouseData* mouse = (WSCmouseData*)_mlist[pos];
    _mlist.del((void*)mouse);
    delete mouse;
  }
}
void WSDmouse::setMouseCursor(long){}
void WSDmouse::setupMouseCursor(char*){}
void WSDmouse::addMouseCursor(long id,long data,char* str){
  WSCmouseData* mouse = NULL;
  long i;
  long num = _mlist._num;
  WSCmouseData** mouse_list = (WSCmouseData**)_mlist._data;
  for(i=0; i<num; i++){
    WSCmouseData* mouse1 = mouse_list[i];
    if (mouse1->id == id){
      mouse1->data = data;
      mouse1->setName(str);
      _mlist.del((void*)mouse1);
      mouse = mouse1;
      break;
    }
  }
  if (mouse == NULL){
    mouse = new WSCmouseData;
    mouse->data = data;
    mouse->id = id;
    mouse->setName(str);
  }
  num = _mlist._num;
  WSCmouseData** mouse_list2 = (WSCmouseData**)_mlist._data;
  for(i=0; i<num; i++){
    WSCmouseData* mouse1 = mouse_list2[i];
    if (mouse1->id < id){
      continue;
    }
    _mlist.add((void*)mouse,i);
    return;
  } 
  _mlist.add((void*)mouse);
}
void WSDmouse::setGlobalMouseHook(WSCbool (*func)(long,WSCbase*)){
  _global_proc = func;
}
void WSDmouse::setSpecialGlobalMouseHook(WSCbool (*func)(long,WSCbase*)){
  _special_global_proc = func;
}
long WSDmouse::getMouseCursorId(long pos){
  long num = _mlist.getNum();
  if (pos < 0){
    return -1;
  }

  if (pos < num){
    WSCmouseData* mouse = (WSCmouseData*)_mlist[pos];
    return mouse->id;
  }
  return -1;
}

char* WSDmouse::getMouseCursorName(long id){
  long num = _mlist.getNum();
  long i;
  WSCmouseData** mouse_list = (WSCmouseData**)_mlist.getBuf();
  for(i=0; i<num; i++){
    WSCmouseData* mouse = mouse_list[i];
    if (mouse->id == id){
      if (mouse->name != NULL){
        return mouse->name;
      }else{
        return "";
      }
    }
  }
  return NULL;
}

long WSDmouse::getMouseCursor(long id){
  if (_global_mouse_id != -1){
    id = _global_mouse_id;
  }
  long num = _mlist.getNum();
  long i;
  WSCmouseData** mouse_list = (WSCmouseData**)_mlist.getBuf();
  for(i=0; i<num; i++){
    WSCmouseData* mouse = mouse_list[i];
    if (mouse->id == id){
      return mouse->data;
    }
  }
  id = 0;
  for(i=0; i<num; i++){
    WSCmouseData* mouse = mouse_list[i];
    if (mouse->id == id){
      return mouse->data;
    }
  }
  if (num > 0){
    WSCmouseData* mouse = (WSCmouseData*)_mlist[0];
    return mouse->data;
  }
  return 0;
}

WSCbool WSDmouse::getMousePosition(short*,short*){
  return False;
}

WSCbool WSDmouse::getMousePosition(short*,short*,WSCbase*){
  return False;
}

void WSDmouse::setMousePosition(short x,short y){}
void WSDmouse::resetMouseCursors(){}

void WSDmouse::addMouseFocusClient(WSCbase* client){
  if (client == NULL){
    return;
  }
  deleteMouseFocusClient(client);
  _mouse_focus_list.add((void*)client,0);
}
void WSDmouse::deleteMouseFocusClient(WSCbase* client){
  int i;
  int num = _mouse_focus_list.getNum();
  for(i=0; i<num; i++){
    if (_mouse_focus_list[i] == (void*)client){
      _mouse_focus_list.delPos(i);
      i--;
      num--;
    }
  }
}
WSCbase* WSDmouse::getMouseFocusClient(){
  if (_mouse_focus_list.getNum() > 0){
    WSCbase* fcl = (WSCbase*)_mouse_focus_list[0];
#ifndef NO_MODAL
    WSCbase* mcl = WSGIappDev()->getModalClient();
    if (mcl != NULL && fcl != NULL){
      WSCbase* pcl = fcl;
      WSCbool fl = False;
      while(1){
        if (pcl == NULL){
          break;
        }
        if (mcl == pcl){
          fl = True;
          break;
        }
        WSDdev* dev = pcl->getowndev();
        if (dev == NULL){
          pcl = pcl->getParent();
          continue;
        }
        WSDdev* pdev = dev->getParentDev();
        if (pdev == NULL){
          break;
        }
        pcl = pdev->getAttachedClient();
      }
      if (fl == False){
        return NULL;
      }
    }
#endif
    return fcl; 
  }
  return NULL;
}
void WSDmouse::setMouseUngrabClient(WSCbase* cl){
  _ungrab_client = cl;
}
WSCbase* WSDmouse::getMouseUngrabClient(){
  return _ungrab_client;
}
void* WSDmouse::getGlobalMouseHook(){
  return (void*)_global_proc;
}
void* WSDmouse::getSpecialGlobalMouseHook(){
  return (void*)_special_global_proc;
}
WSCbool WSDmouse::execGlobalMouseHook(long ev,WSCbase* base){
  if (_special_global_proc != NULL){
    return _special_global_proc(ev,base);
  }
  if (_global_proc != NULL){
    return _global_proc(ev,base);
  }
  return True;
}
void WSDmouse::resetGlobalMouseId(){
  _global_mouse_id = -1;
}
void WSDmouse::setGlobalMouseId(short id){
  _global_mouse_id = id;
}
short WSDmouse::getGlobalMouseId(){
  return _global_mouse_id;
}
WSCbool WSDmouse::getGlobalMouseIdStatus(){
  if (_global_mouse_id == -1){
    return False;
  }else{
    return True;
  }
}
void WSDmouse::setGlobalMouseSensitive(WSCbool fl){
  _global_mouse_sensitive = fl;
}
WSCbool WSDmouse::getGlobalMouseSensitive(){
  return _global_mouse_sensitive;
}
#ifndef NO_VIRTUAL_EVENT_SEND
long WSDmouse::onMousePress(short x,short y,long btn){
  return WS_NO_ERR;
}
long WSDmouse::onMouseRelease(short x,short y,long btn){
  return WS_NO_ERR;
}
long WSDmouse::onMouseMove(short x,short y,long btn){
  return WS_NO_ERR;
}
#endif
