//
// Copyright (C) 1999-2009 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#include <WSDnwDev.h>
#include <WSDdraw.h>
#include <WSCbase.h>

WSMFclassInit(WSDnwDev,WSDdev);

WSDnwDev::WSDnwDev(){
  _use_pixmap_ = False;
  _no_child = True;
  _clipping = True;
  _clip_children = True;
}

short     WSDnwDev::_copy_pixmap_x = 0;
short     WSDnwDev::_copy_pixmap_y = 0;
WSCushort WSDnwDev::_copy_pixmap_w = 0;
WSCushort WSDnwDev::_copy_pixmap_h = 0;

WSCushort WSDnwDev::_pixmap_w = 0;
WSCushort WSDnwDev::_pixmap_h = 0;
short WSDnwDev::_back_x = 0;
short WSDnwDev::_back_y = 0;

void WSDnwDev::getBackXY(short* x,short* y){
  *x = _back_x;
  *y = _back_y;
}
void WSDnwDev::getPixmapWH(WSCushort* h,WSCushort* w){
  *w = _pixmap_w;
  *h = _pixmap_h;
}
void WSDnwDev::getCopyPixmapGeom(short* x,short* y,WSCushort* w,WSCushort* h){
  *x = _copy_pixmap_x;
  *y = _copy_pixmap_y;
  *w = _copy_pixmap_w;
  *h = _copy_pixmap_h;
}

long WSDnwDev::setValue(long kind,void* data){
  char val;
//  WSCushort sval;
  switch(kind){
    case WSDEV_X:{
      _w_x = *(short*)data;
      WSCbase* client = getAttachedClient();
      if (client != NULL){
        WSClistData* children = client->getChildrenPtr();
        if (children != NULL){
          long num = children->getNum();
          if (num > 0){
            int i;
            for(i=0; i<num; i++){
              WSCbase* child = (WSCbase*)children->getData(i);
              child->setGeometryChanged(True);
              child->needUpdate();
              child->update();
            }
            break;
          }
        }
      }
      break;
    }
    case WSDEV_Y:{
      _w_y = *(short*)data;
      WSCbase* client = getAttachedClient();
      if (client != NULL){
        WSClistData* children = client->getChildrenPtr();
        if (children != NULL){
          long num = children->getNum();
          if (num > 0){
            int i;
            for(i=0; i<num; i++){
              WSCbase* child = (WSCbase*)children->getData(i);
              child->setGeometryChanged(True);
              child->needUpdate();
              child->update();
            }
            break;
          }
        }
      }
      break;
    }
    case WSDEV_USE_PIXMAP:
      val = *(char*)data;
      setUsePixmap(val);
      break;
#ifndef NO_SCALE_OFFSET
    case WSDEV_SCALE_OFFSET:
      setScalePtr((double*)data);
      break;
    case WSDEV_X_OFFSET:
      setXOffsetPtr((short*)data);
      break;
    case WSDEV_Y_OFFSET:
      setYOffsetPtr((short*)data);
      break;
#endif
    case WSDEV_CLIPPING:
      _clipping = *(WSCbool*)data;
      break;
    case WSDEV_CLIPCHILDREN:
      _clip_children = *(WSCbool*)data;
      break;
  }
  
  return WS_NO_ERR;
}

WSCbool WSDnwDev::getReady(){
//  WSDdev* pt = getParentDev();
  WSDdev* pt = getEventParentDev();
  if (pt == NULL){
    return False;
  }
#if 0 //optimizing..
  long d1 = getDeviceResource();
  long d2 = getWindowResource();
  long d3 = getContextResource();
  if (d1 == -1 || d2 == -1 || d3 == -1){
    return False;
  }
#else
  if ( pt->getDeviceResource() == -1){
    return False;
  }
  if ( pt->getWindowResource() == -1){
    return False;
  }
  if ( pt->getContextResource() == -1){
    return False;
  }
#endif
  return True;
}
long WSDnwDev::getWindowResource(){
//  WSDdev* pt = getParentDev();
  WSDdev* pt = getEventParentDev();
  if (pt == NULL){
    return -1;
  }
  return pt->getWindowResource(); 
}
long WSDnwDev::getContextResource(){
//  WSDdev* pt = getParentDev();
  WSDdev* pt = getEventParentDev();
  if (pt == NULL){
    return -1;
  }
  return pt->getContextResource(); 
}
long WSDnwDev::getSpecialResource(){
//  WSDdev* pt = getParentDev();
  WSDdev* pt = getEventParentDev();
  if (pt == NULL){
    return WS_ERR;
  }
  return pt->getSpecialResource(); 
}
long WSDnwDev::getDeviceResource(){
//  WSDdev* pt = getParentDev();
  WSDdev* pt = getEventParentDev();
  if (pt == NULL){
    return -1;
  }
  return pt->getDeviceResource(); 
}
WSCbool WSDnwDev::isAvailableEvent(long ev_kind){
  long val = WSGFeventToMaskBit(ev_kind);
  if ( val & getAvailableEventBit() ){
    return True;
  }
  return False;
}

long WSDnwDev::beginDraw(short x,short y,WSCushort w,WSCushort h,WSCbool absolute,WSCbool scaling){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    return WS_ERR;
  }
  //getReady() calls getWindow() or getPixmap() 
  //getPixmap() creates pixmap with _w,_h that has no effect values
  _h = 1;
  _w = 1;
  if (getReady() == False){
    return WS_ERR;
  }

  _absolute_draw = absolute;
  if (scaling != False){
    double scale = getScale();
    short xoff = 0;
    short yoff = 0;
    getOffset(&xoff,&yoff);
    x = (short)((x + xoff)*scale);
    y = (short)((y + yoff)*scale);
    w = (WSCushort)(w * scale);
    h = (WSCushort)(h * scale);
  }

  _expose_sequence_ch = False;
  _draw_pixmap_only_ch = False;
  WSDdev* pdev = getEventParentDev();
  if ( pdev == NULL ){
    return WS_ERR;
  }

  long tx = 0;
  long ty = 0;
  WSCulong tw = 0;
  WSCulong th = 0;
  getClippedArea(x,y,w,h,&tx,&ty,&tw,&th);
#if 0
  x = tx;
  y = ty;
  w = tw;
  h = th;
#endif
#if 0
  WSDdev* ppdev = getParentDev();
  WSDdev* _this = this;
  while(1){
    if (_this->getIsNonWindow() == False){
      break;
    }
    if (ppdev->getIsNonWindow() != False &&
        ppdev->clipChildren() != False &&
        _this->clipping() != False){
      long cx = ppdev->_w_x;
      long cy = ppdev->_w_y;
      WSCbase* pclient = ppdev->getAttachedClient();
      WSCulong cw = pclient->getProperty(WSNwidth);
      WSCulong ch = pclient->getProperty(WSNheight);

      double pscale = ppdev->getScale();
      short xoff = 0;
      short yoff = 0;
      ppdev->getOffset(&xoff,&yoff);
      cx = (short)((cx + xoff)*pscale);
      cy = (short)((cy + yoff)*pscale);
      cw = (WSCushort)(cw * pscale);
      ch = (WSCushort)(ch * pscale);

      long px,py;
      WSCulong pw,ph;
      WSGFandArea(x,y,w,h,cx,cy,cw,ch,&px,&py,&pw,&ph);
      x = px;
      y = py;
      w = pw;
      h = ph;
    }
    if (ppdev == NULL){
      break;
    }
    _this = ppdev;
    ppdev = ppdev->getParentDev();
    if (ppdev == NULL || ppdev->getIsNonWindow() == False){
      break;
    }
  }
#endif
#if 0 //optimizing..
  if (_absolute_draw != False){
    void* wdev = pdev->cast("WSDwindowDev");
    if (wdev != NULL){
      WSCuchar fl;
      pdev->getValue(WSDEV_EXPOSE_SEQUENCE,(void*)&fl);
      _expose_sequence_bak = fl;
      if (_expose_sequence_bak != False){
        fl = False;
        pdev->setValue(WSDEV_EXPOSE_SEQUENCE,(void*)&fl);
        _expose_sequence_ch = True;
      }
    }
  }
  pdev->beginDraw( x, y, w, h,absolute,False);
#else
  if (_absolute_draw != False){
    _expose_sequence_bak = pdev->getExposeSequence();
    if (_expose_sequence_bak != False){
      pdev->setExposeSequence(False);
      _expose_sequence_ch = True;
    }
  }
  if (_draw_pixmap_only != False){
    _draw_pixmap_only_bk = pdev->getDrawPixmapOnly();
    if (_draw_pixmap_only_bk == False){
      pdev->setDrawPixmapOnly(True);
      _draw_pixmap_only_ch = True;
    }
  }
#ifndef MSW
  pdev->internalBeginDraw( x, y, w, h,absolute,False);
#else
  pdev->beginDraw( x, y, w, h,absolute,False);
#endif
#endif
  WSCbool use_pix = getUsePixmap();
  if (use_pix == False){
    _x = x; 
    _y = y; 
  }else{
    _back_x = x;
    _back_y = y;
    _x = 0;
    _y = 0;
  }
  _w = w; 
  _h = h; 

  short ex,ey;
  WSCushort ew,eh;
  long ret =   getExposedArea(&ex,&ey,&ew,&eh,False);
  if (ret != WS_NO_ERR){
    return WS_ERR;
  }
  draw->beginDraw(this,getDeviceResource(),getWindowResource(),getContextResource());
#if 0
  short ex2 = ex + ew < x + w ? ex + ew : x + w; 
  short ey2 = ey + eh < y + h ? ey + eh : y + h; 
  ex = ex > x ? ex : x; 
  ey = ey > y ? ey : y; 
#endif
  short ex2 = ex + ew < tx + tw ? ex + ew : tx + tw; 
  short ey2 = ey + eh < ty + th ? ey + eh : ty + th; 
  ex = ex > tx ? ex : tx; 
  ey = ey > ty ? ey : ty; 
  if (ex2 > ex){
    ew = ex2 - ex;
  }else{
    ew = 0;
  }
  if (ey2 > ey){
    eh = ey2 - ey;
  }else{
    eh = 0;
  }
  if (use_pix == False){
    if (absolute == False){
      draw->setRegion(ex,ey,ew,eh);
    }else{
#if 0
      draw->setRegion(x,y,w,h);
#endif
      draw->setRegion(tx,ty,tw,th);
    }
  }else{
    if (absolute == False){
      _copy_pixmap_x = ex - x;
      _copy_pixmap_y = ey - y;
      _copy_pixmap_w = ew;
      _copy_pixmap_h = eh;
      draw->setRegion(0,0,w,h);
    }else{
      _copy_pixmap_x = 0;
      _copy_pixmap_y = 0;
      _copy_pixmap_w = w;
      _copy_pixmap_h = h;
      draw->setRegion(0,0,w,h);
    }
  }

  return WS_NO_ERR;
}

long WSDnwDev::endDraw(){
  WSDdraw* draw = getDraw();
  if (draw == NULL){
    _absolute_draw = False;
    return WS_ERR;
  }
  draw->endDraw();

  if (getUsePixmap() != False){
    _copy_pixmap_to_window();
  }

//  WSDdev* pdev = getParentDev();
  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){
#ifndef MSW
    pdev->internalEndDraw();
#else
    pdev->endDraw();
#endif
    if (_absolute_draw != False && _expose_sequence_ch != False ){
//      void* wdev = pdev->cast("WSDwindowDev");
//      if (wdev != NULL){
//	    WSCuchar fl = _expose_sequence_bak;
//        pdev->setValue(WSDEV_EXPOSE_SEQUENCE,(void*)&fl);
        pdev->setExposeSequence(_expose_sequence_bak);
//      }
    }
    if (_draw_pixmap_only_ch != False){
        pdev->setDrawPixmapOnly(_draw_pixmap_only_bk);
    }
  } 
  _absolute_draw = False;
  return WS_NO_ERR;
}
long WSDnwDev::exposeArea(long x,long y,WSCulong w,WSCulong h,WSCbool scaling){

  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){
    WSCbool scaled = getScaled();
    short xoff = 0;
    short yoff = 0;
    getOffset(&xoff,&yoff);
    if (scaled != False && scaling != False){
      double scale = getScale();
      pdev->exposeArea((short)((x + xoff) * scale),
                        (short)((y + yoff) * scale),
                        (WSCushort)(w * scale),
                        (WSCushort)(h * scale),False);
    }else{
      pdev->exposeArea(x + xoff,y + yoff,w,h,False);
    }
  }
  return WS_NO_ERR;
}

void WSDnwDev::onExpose(WSCrect* area){

  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){
    WSCbool scaled = getScaled();
    short xoff = 0;
    short yoff = 0;
    getOffset(&xoff,&yoff);
    if (scaled != False){
      double scale = getScale();
      WSCrect scaled_area;
      scaled_area.x = (long)((area->x + xoff) * scale);
      scaled_area.y = (long)((area->y + yoff) * scale);
      scaled_area.width = (WSCulong)(area->width * scale);
      scaled_area.height = (WSCulong)(area->height * scale);

      pdev->onExpose(&scaled_area);
    }else{
      WSCrect area2;
      area2.x = area->x + xoff;
      area2.y = area->y + yoff;
      area2.width = area->width;
      area2.height = area->height;
      pdev->onExpose(&area2);
    }
  }
}
long WSDnwDev::getClippedArea(long x,long y,WSCulong w,WSCulong h,
                              long* ox,long* oy,WSCulong* ow,WSCulong* oh){
  WSDdev* ppdev = getParentDev();
  WSDdev* _this = this;
  if (ppdev == NULL){
    *ox = x;
    *oy = y;
    *ow = w;
    *oh = h;
    return WS_NO_ERR;
  }
  while(1){
    if (_this->getIsNonWindow() == False){
      break;
    }
    if (ppdev->getIsNonWindow() != False &&
        ppdev->clipChildren() != False &&
        _this->clipping() != False){

      long cx = ppdev->_w_x;
      long cy = ppdev->_w_y;
      WSCbase* pclient = ppdev->getAttachedClient();
      WSCulong cw = pclient->getProperty(WSNwidth);
      WSCulong ch = pclient->getProperty(WSNheight);

      double pscale = ppdev->getScale();
      short xoff = 0;
      short yoff = 0;
      ppdev->getOffset(&xoff,&yoff);
      cx = (short)((cx + xoff)*pscale);
      cy = (short)((cy + yoff)*pscale);
      cw = (WSCushort)(cw * pscale);
      ch = (WSCushort)(ch * pscale);

      long px,py;
      WSCulong pw,ph;
      WSGFandArea(x,y,w,h,cx,cy,cw,ch,&px,&py,&pw,&ph);
      x = px;
      y = py;
      w = pw;
      h = ph;
    }
    if (ppdev == NULL){
      break;
    }
    _this = ppdev;
    ppdev = ppdev->getParentDev();
    if (ppdev == NULL || ppdev->getIsNonWindow() == False){
      break;
    }
  }
  *ox = x; 
  *oy = y; 
  *ow = w; 
  *oh = h; 
  return WS_NO_ERR;
}
long WSDnwDev::clearArea(long x,long y,WSCulong w,WSCulong h,WSCbool create_event,WSCbool scaling){

  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){
    WSCuchar back;
    if (_use_pixmap_ != False){
      pdev->getValue(WSDEV_PIXMAP_STYLE,(void*)&back);
      if (back != WS_DYNAMIC_PIXMAP){
        WSCuchar fl = WS_DYNAMIC_PIXMAP;
        pdev->setValue(WSDEV_PIXMAP_STYLE,(void*)&fl);
      }
    }

    WSCbool scaled = getScaled();
    short xoff = 0;
    short yoff = 0;
    getOffset(&xoff,&yoff);
    if (scaled != False && scaling != False){
      double scale = getScale();
      x = ((x + xoff) * scale);
      y = ((y + yoff) * scale);
      w = (w * scale);
      h = (h * scale);
    }else{
      x += xoff;
      y += yoff;
    }

    getClippedArea(x,y,w,h,&x,&y,&w,&h);

    if (w > 0 && h > 0){
      pdev->clearArea(x,y,w,h, create_event,False);
    }

#if 0
    if (scaled != False && scaling != False){
      double scale = getScale();
      pdev->clearArea((short)((x + getXOffset()) * scale),
                        (short)((y + getYOffset()) * scale),
                        (WSCushort)(w * scale),
                        (WSCushort)(h * scale),create_event,False);
    }else{
      pdev->clearArea(x + getXOffset(),y + getYOffset(),w,h,
                      create_event,False);
    }
#endif
    if (_use_pixmap_ != False){
      if (back != WS_DYNAMIC_PIXMAP){
        pdev->setValue(WSDEV_PIXMAP_STYLE,(void*)&back);
      }
    }
  }
  return WS_NO_ERR;
}
long WSDnwDev::erase(long x,long y,WSCulong w,WSCulong h,WSCbool create_event,WSCbool scaling){

  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){
//    WSCuchar back;
    WSCbool scaled = getScaled();
    short xoff = 0;
    short yoff = 0;
    getOffset(&xoff,&yoff);
    if (scaled != False && scaling != False){
      double scale = getScale();
      pdev->erase((short)((x + xoff) * scale),
                        (short)((y + yoff) * scale),
                        (WSCushort)(w * scale),
                        (WSCushort)(h * scale),create_event,False);
    }else{
      pdev->erase(x + xoff,y + yoff,w,h,create_event,False);
    }
  }
  return WS_NO_ERR;
}
long WSDnwDev::copyArea(long x,long y,WSCulong w,WSCulong h,long dx,long dy){
  WSDdev* pdev = getEventParentDev();
  if ( pdev != NULL ){
    WSCbool scaled = getScaled();
    short xoff = 0;
    short yoff = 0;
    getOffset(&xoff,&yoff);
    if (scaled != False){
      double scale = getScale();
      pdev->copyArea((short)((x + xoff) * scale),
                        (short)((y + yoff) * scale),
                        (WSCushort)(w * scale),
                        (WSCushort)(h * scale),
                        (short)((dx + xoff) * scale),
                        (short)((dy + yoff) * scale));
    }else{
      pdev->copyArea(x + xoff,y + yoff,w,h, dx + xoff,dy + yoff );
    }
  }
  return WS_NO_ERR;
}

void WSDnwDev::getDispAddr(short* ox,short* oy){
   short cx = 0,cy = 0;
   WSCbase* client = getAttachedClient();
   if (client == NULL){
     *ox = 0;
     *oy = 0;
     return;
   }

   if (client->existProperty(WSNx) == True){
     client->getPropertyV(WSNx,&cx);
   }
   if (client->existProperty(WSNy) == True){
     client->getPropertyV(WSNy,&cy);
   }

   short px = 0,py = 0;
   WSDdev* parent = getParentDev();
   if (parent != NULL){
     parent->getDispAddr(&px,&py);
   }
   short xoff = 0;
   short yoff = 0;
   getOffset(&xoff,&yoff);
   short cx2 = (short)(cx + xoff);
   short cy2 = (short)(cy + yoff);
   if (getScaled() != False){
     double scale = getScale();
     cx2 = (short)(cx2 * scale);
     cy2 = (short)(cy2 * scale);
   }
   short absolute_cx;
   short absolute_cy;

   if (parent != NULL){
     parent->getAbsoluteAddr(cx2,cy2,&absolute_cx,&absolute_cy);
   }
   *ox = px + absolute_cx;
   *oy = py + absolute_cy;
   return;
}

void WSDnwDev::getAbsoluteAddr(short x,short y,short* ox,short* oy){
  WSDdev* parent = getParentDev();
  *ox =0;
  *oy =0;
  if (parent != NULL){
    parent->getAbsoluteAddr(x,y,ox,oy);
  }
  short xoff = 0;
  short yoff = 0;
  getOffset(&xoff,&yoff);
  *ox = (*ox + xoff);
  *oy = (*oy + yoff);
  if (getScaled() != False){
    double scale = getScale();
    *ox = (short)(*ox * scale);
    *oy = (short)(*oy * scale);
  }
}

long WSDnwDev::setAbsoluteAddrChangeHandler(void(*hd)(WSDdev*,void*),void* data){
  WSDdev* parent = getParentDev();
  if (parent != NULL){
    parent->setAbsoluteAddrChangeHandler(hd,data);
  }
  return WS_NO_ERR;
}

WSCbool WSDnwDev::getUsePixmap(){
  if (_use_pixmap_ == False){
    return _use_pixmap_;
  }else{
    WSDdev* parent = getParentDev();
//    if (parent == NULL){
//      return _use_pixmap_;
//    }
//    void* wdev = parent->cast("WSDwindowDev");
//    if (wdev != NULL ){
//	  WSCuchar fl = 0;
//	  parent->getValue(WSDEV_PIXMAP_STYLE,(void*)&fl);
//    if (fl != WS_DIRECT_WINDOW){
//      return False;
//    }
    if (parent != NULL && parent->getPixmapStyle() != WS_DIRECT_WINDOW){
      return False;
    }
    return _use_pixmap_;
  }
}

void WSDnwDev::setUsePixmap(WSCbool fl){
  _use_pixmap_ = fl;
}


long WSDnwDev::setEventOrder(WSDdev* dev,char direction){
//  WSDdev* parent = getParentDev();
  WSDdev* parent = getEventParentDev();
  if (parent == NULL){
    return WS_ERR;
  }
  return parent->setEventOrder(dev,direction); 
}

long WSDnwDev::setEvent(WSDdev* dev,WSCbool on_off){
//  WSDdev* parent = getParentDev();
  WSDdev* parent = getEventParentDev();
  if (parent == NULL){
    return WS_ERR;
  }
  return parent->setEvent(dev,on_off); 
}

long WSDnwDev::cancelEvent(WSDdev* dev){
  if (getEventRegistered() == False){
    return WS_NO_ERR;
  }
  if (dev == this){
    setEventRegistered(False);
  }

//  WSDdev* parent = getParentDev();
  WSDdev* parent = getEventParentDev();
  if (parent == NULL){
    return WS_ERR;
  }
  parent->cancelEvent(dev); 
  return WS_NO_ERR;
}

long WSDnwDev::setEnableEventBit(long ev_bit){
  if (getEventRegistered() != False){
    if ( ev_bit & getAvailableEventBit() ){
      long evmask = getEventMaskBit();
      evmask |= ( ev_bit & getAvailableEventBit() ); 
      setEventMaskBit( evmask );
    }
    return WS_NO_ERR;
  }
//  WSDdev* parent = getParentDev();
  WSDdev* parent = getEventParentDev();
  if (parent == NULL){
    return WS_ERR;
  }
  if ( ev_bit & getAvailableEventBit() ){
    parent->setEvent(this,True);
    setEventMaskBit(  ev_bit & getAvailableEventBit() );
    setEventRegistered(True);
    return WS_NO_ERR;
  }
  return WS_ERR;
}

long WSDnwDev::setEnableEvent(long ev_kind){
  if (getEventRegistered() != False){
    if (isAvailableEvent(ev_kind) != False){
      long evmask = getEventMaskBit();
      evmask |= WSGFeventToMaskBit(ev_kind); 
      setEventMaskBit(evmask);
    }
    return WS_NO_ERR;
  }

//  WSDdev* parent = getParentDev();
  WSDdev* parent = getEventParentDev();
  if (parent == NULL){
    return WS_ERR;
  }

  if (isAvailableEvent(ev_kind) != False){
    setEventRegistered(True);
    setEventMaskBit(WSGFeventToMaskBit(ev_kind));
    parent->setEvent(this,True);
    return WS_NO_ERR;
  }
  return WS_ERR;
}
long WSDnwDev::setDisableEvent(long ev_kind){
  long mask = getEventMaskBit();
  mask &= ~WSGFeventToMaskBit(ev_kind);
  setEventMaskBit(mask);

  WSDdev* parent = getParentDev();
//  WSDdev* parent = getEventParentDev();
  if (parent == NULL){
    return WS_ERR;
  }

  if (mask == 0){
    parent->setEvent(this,False);
    return WS_NO_ERR;
  }

  return WS_ERR;
}

long WSDnwDev::getAvailableEventBit(){
   return  ( WSEV_MOUSE_MOVE_BIT | WSEV_EXPOSE_BIT |
       WSEV_RESIZE_BIT    | WSEV_MOUSE_IN_BIT  |
       WSEV_MOUSE_OUT_BIT | WSEV_MOUSE_PRESS_BIT |
       WSEV_FOCUS_CH_BIT  | WSEV_MOUSE_RELEASE_BIT |
       WSEV_INITIALIZE_BIT | WSEV_VISIBLE_CH_BIT  |
       WSEV_KEY_PRESS_BIT  | WSEV_KEY_RELEASE_BIT |
       WSEV_KEY_HOOK_BIT );  
}
WSCbool WSDnwDev::getIsNonWindow(){
  return True;
}
WSCbool WSDnwDev::clipping(){
  return _clipping;
}
WSCbool WSDnwDev::clipChildren(){
  return _clip_children;
}
