//
// Copyright (C) 1999-2008 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#ifndef NO_SOCKET
#ifndef NO_THREAD
#include <WSCvmultiServerSocket.h>
#include <WSCclassInformation.h>
#include <WSCdevice.h>
#include <WSCblink.h>
#include <WSCcolorSet.h>
#include <WSCimageSet.h>
#include <WSDdev.h>
#include <WSDcolor.h>
#include <WSCfontSet.h>
#include <WSDappDev.h>

WSMFguiClassInitialize(WSCvmultiServerSocket, WSCngbase);
WSMFversion(WSCvmultiServerSocket, WSCngbase);

WSCvmultiServerSocket::WSCvmultiServerSocket(WSCbase* base, char* objname):
			WSCngbase(base, objname){
  _socket = NULL;
  _accept_thread = NULL;
  _addr = WSGFstrdup("");
  _port = 9000;
  _on = False;
  _timeout = 60;
  _title_string.setString("Multi\nServer\nSocket");
  _disable = False;

  WSMFpropertyCreateStart
    WSMFparentCheckVerSrc(WSCvmultiServerSocket);

    WSMFpropertyCreate(WSNip,      char*,   _addr,WSSip);
    WSMFpropertyCreate(WSNport,    WSCushort,   _port,WSSport);
    WSMFpropertyCreate(WSNtimeout, WSCulong,   _timeout,WSStimeout);
    WSMFpropertyCreate(WSNrunning, WSCbool,   _on,WSSrunning);
      WSMFpropertySetSelection(WSRbool3,WSRbool3D);

    WSMFaddTrigger(WSEV_ACTIVATE  );
  WSMFpropertyCreateEnd

}
WSMFproperty( WSCvmultiServerSocket, WSNip, char*, _addr ,WSGFstrdup(""));
WSMFproperty( WSCvmultiServerSocket, WSNport, WSCushort, _port ,9000);
WSMFproperty( WSCvmultiServerSocket, WSNrunning, WSCbool, _on ,False);
WSMFproperty( WSCvmultiServerSocket, WSNtimeout, WSCulong, _timeout ,60);

void WSCvmultiServerSocket::setWorkWSNip(char* ip){
  if (_socket == NULL){
    _socket = WSDsocket::getNewInstance();
  }
  _socket->setAddr(ip); 
};
void WSCvmultiServerSocket::getWorkWSNip(char**){};

void WSCvmultiServerSocket::setWorkWSNport(WSCushort port){
  if (_socket == NULL){
    _socket = WSDsocket::getNewInstance();
  }
  _socket->setPort(port); 
};
void WSCvmultiServerSocket::getWorkWSNport(WSCushort*){};

void WSCvmultiServerSocket::setWorkWSNrunning(WSCbool fl){
  if (_socket == NULL){
    _socket = WSDsocket::getNewInstance();
  }
#ifndef WS_EMBED
  if (getPropertyEditMode() != False){
    return;
  }
#endif
  if (fl == False){
    if (_socket->getAcceptStarted() != False){
//      _socket->stopAcceptEx();
      _socket->close();
      _error_str = _socket->getLastError();
    }
  }else{
    if (_socket->getAcceptStarted() == False){
      _socket->setAddr(_addr); 
      _socket->setPort(_port); 
      _socket->setClTimeout(_timeout);
//      _socket->setUserData(this);
//      _socket->acceptEx(_accepted_work);

      if (_disable == False){
        if (_accept_thread == NULL){
          _accept_thread = WSDthread::getNewInstance();
          _accept_thread->setFunction(_accept_thread_proc);
        }
        if (_accept_thread != NULL){
          _accept_thread->createThread(this);
          _accept_thread->detachThread();
        }
        _error_str = _socket->getLastError();
      }
    }
  }
};
void WSCvmultiServerSocket::getWorkWSNrunning(WSCbool*){};

void WSCvmultiServerSocket::setWorkWSNtimeout(WSCulong timeout){
  if (_socket == NULL){
    _socket = WSDsocket::getNewInstance();
  }
  _socket->setClTimeout(timeout);
};
void WSCvmultiServerSocket::getWorkWSNtimeout(WSCulong*){};



void WSCvmultiServerSocket::onActivate(){
  execProcedure(WSEV_ACTIVATE);
}

WSCvmultiServerSocket::~WSCvmultiServerSocket() {
  if (_socket != NULL){
    if (_socket->getAcceptStarted() != False){
      _socket->stopAcceptEx();
    }
    delete _socket;
    _socket = NULL;
  }

  if (_addr != NULL){
    delete[] _addr;
    _addr = NULL;
  }
}

void* WSCvmultiServerSocket::_accept_thread_proc(WSDthread* thr, void* ptr){
  WSCvmultiServerSocket* _this = (WSCvmultiServerSocket*)ptr;
  while(1){
    void* sock = _this->_socket->accept();
    if (sock == NULL){
      return NULL;
    }
    WSCstring dest_addr = _this->_socket->getDestAddr();
    WSCvmultiServerSocket* ss = new WSCvmultiServerSocket(NULL,"");
    ss->initialize();
    ss->_disable = True;
    ss->_handle = sock;
    ss->_socket = WSDsocket::getNewInstance();
    ss->_socket->setAddr(_this->_addr);
    ss->_socket->setPort(_this->_port);
    ss->_socket->setClTimeout(_this->_timeout);
    ss->_socket->setDestAddr(dest_addr);
    WSClistData* procedures = _this->getProcedureList();
    int i;
    int num = procedures->getNum();
    for(i=0; i<num; i++){
      WSCprocedure* proc = (WSCprocedure*)procedures->getData(0);
      ss->addProcedureV(proc->getProcName(),proc->getFunctionName(),
                        (void(*)(WSCbase*)) proc->getProc(),proc->getTrigger());
    }
    WSDthread* server_thread = WSDthread::getNewInstance();
    if (server_thread != NULL){
      server_thread->setFunction(_server_thread_proc);
      server_thread->setDestroyFlag(True);
      server_thread->createThread(ss);
      server_thread->detachThread();
    }else{
      delete ss;
    }
  }
}
void* WSCvmultiServerSocket::_server_thread_proc(WSDthread* thr, void* ptr){
  WSCvmultiServerSocket* ss = (WSCvmultiServerSocket*)ptr;
  ss->execProcedure(WSEV_ACTIVATE);
  delete ss;
  return NULL;
}
long WSCvmultiServerSocket::read(WSCuchar* buf,long size){
  if (_socket == NULL){
    return -1;
  }
  if (_handle == NULL){
    return -1;
  }
  long ret = _socket->read(_handle,buf,size);
  _error_str = _socket->getLastError();
  return ret;
}
long WSCvmultiServerSocket::write(WSCuchar* buf,long size){
  if (_socket == NULL){
    return -1;
  }
  if (_handle == NULL){
    return -1;
  }
  long ret = _socket->write(_handle,buf,size);
  _error_str = _socket->getLastError();
  return ret;
}
WSCstring WSCvmultiServerSocket::getDestAddr(){
  if (_socket == NULL){
    WSCstring tmp;
    return tmp;
  }
  return _socket->getDestAddr();
}
WSDsocket* WSCvmultiServerSocket::getSocket(){
  return _socket;
}
void WSCvmultiServerSocket::getErrorMsg(char* buffer,long size){
  if (size < 1){
    return;
  }
  strncpy(buffer,(char*)_error_str,size);
  buffer[size-1] = 0;
}
WSCstring WSCvmultiServerSocket::getErrorMessage(){
  WSCstring ret;
  ret.setString(_error_str);
  return ret;
}
#endif
#endif
