//
// Copyright (C) 1999-2009 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#include <WSCvform.h>
#include <WSCclassInformation.h>
#include <WSCdevice.h>
#include <WSDdev.h>
#include <WSDappDev.h>
#include <WSCbaseList.h>
#include <WSDwindowDev.h>
#include <WSCguiPolicy.h>

WSMFguiClassInitialize(WSCvform,WSCbase);
WSMFversion(WSCvform,WSCbase);
WSMFdefineUseDevice(WSCvform,nwDev);

WSCvform::WSCvform(WSCbase* base,char* objname):WSCbase(base,objname){
  _user_string = NULL;
  _user_value  = 0;
  _x   = 100;
  _y   = 100;
  _w   = 100;
  _h   = 30;
  _shadow_thick = 2;
  _fore_color = WS_DF_FORECOLOR;
  _back_color = WS_DF_BACKCOLOR;
  _ts_color = WS_DF_TOPSHADOWCOLOR;
  _bs_color = WS_DF_BOTTOMSHADOWCOLOR;
  _bp_color = -1;
  _shadow_type = WS_SHADOW_OUT;
  _sensitive = True;
  _sensitive_work = True;
  _no_refresh = False;
#ifndef NO_MOUSE_PROPERTY
  _mouse_no     = 0;
#endif
#ifndef NO_GRADATION
  _grad_no     = 0;
  _grad_margin   = 0;
#endif
  _update_resize   = 0;
#ifndef NO_ANCHOR
  _anchor_top = 0;
  _anchor_bottom= 0;
  _anchor_left = 0;
  _anchor_right = 0;
  _anchor_top_use = False;
  _anchor_bottom_use = False;
  _anchor_left_use = False;
  _anchor_right_use = False;
#endif
  _a1_color = -1;
  _a2_color = -1;
  _a3_color = -1;
  _children._seg_size = 8;
  _mouse_enter = False;
  _no_ext_gui_proc = False;
  _event_mask |= WSEV_EXPOSE_BIT | WSEV_MOUSE_MOVE_BIT | WSEV_MOUSE_IN_BIT | WSEV_MOUSE_OUT_BIT;

  WSGIappObjectList()->addEvent(this,WSEV_GUI_POLICY_CH);

  WSMFpropertyCreateStart
    WSMFparentCheckVerSrc(WSCvform);
    WSMFpropertyCreate(WSNname,  char*, _instance_name,WSSname );
    WSMFpropertyCreate(WSNuserString,  char*, _user_string,WSSuserString );
    WSMFpropertyCreate(WSNuserValue,   long,  _user_value,WSSuserValue  );
    WSMFpropertyCreate(WSNx,           short,   _x ,WSSx );
    WSMFpropertyCreate(WSNy,           short,   _y ,WSSy );
    WSMFpropertyCreate(WSNwidth,        WSCushort, _w ,WSSwidth );
    WSMFpropertyCreate(WSNheight,       WSCushort, _h ,WSSheight );
    WSMFpropertyCreate(WSNshadowThickness,  WSCuchar, _shadow_thick,WSSshadowThickness);
    WSMFpropertyCreate(WSNforeColor,    short,   _fore_color,WSSforeColor);
    WSMFpropertyCreate(WSNbackColor,     short,  _back_color,WSSbackColor);
    WSMFpropertyCreate(WSNtopShadowColor,short,  _ts_color,WSStopShadowColor);
    WSMFpropertyCreate(WSNbottomShadowColor,short, _bs_color,WSSbottomShadowColor );
#ifndef NO_IMAGE
    WSMFpropertyCreate(WSNbackPixmap, short,  _bp_color,WSSbackPixmap );
#endif
    WSMFpropertyCreate(WSNshadowType,  char,  _shadow_type,WSSshadowType );
      WSMFpropertySetSelection(WSRshadowType,WSRshadowTypeD);
    WSMFpropertyCreate(WSNvis,       WSCbool,      _vis,WSSvis );
      WSMFpropertySetSelection(WSRbool2,WSRbool2D);
    WSMFpropertyCreate(WSNdet,              WSCbool,  _sensitive_work,WSSdet);
      WSMFpropertySetSelection(WSRbool1,WSRbool1D);
#ifndef NO_MOUSE_PROPERTY
    WSMFpropertyCreate( WSNmouse, WSCushort,   _mouse_no,WSSmouse  );
#endif
#ifndef NO_GRADATION
    WSMFpropertyCreate(WSNgradation,  WSCuchar,  _grad_no,WSSgradation);
      WSMFpropertySetSelection(WSRgradation,WSRgradationD);
    WSMFpropertyCreate(WSNgradationMargin,WSCuchar,_grad_margin,WSSgradationMargin);
#endif

#ifndef NO_ANCHOR
    WSMFpropertyCreate(WSNanchorTop, WSCushort, _anchor_top,WSSanchorTop);
    WSMFpropertyCreate(WSNanchorBottom,WSCushort,_anchor_bottom,WSSanchorBottom);
    WSMFpropertyCreate(WSNanchorLeft,WSCushort,_anchor_left,WSSanchorLeft);
    WSMFpropertyCreate(WSNanchorRight,WSCushort,_anchor_right,WSSanchorRight);

    WSMFpropertyCreate(WSNanchorTopFlag, WSCbool, _anchor_top_use,WSSanchorTopFlag);
    WSMFpropertySetSelection(WSRbool3, WSRbool3D);
    WSMFpropertyCreate(WSNanchorBottomFlag,WSCbool,_anchor_bottom_use,WSSanchorBottomFlag);
    WSMFpropertySetSelection(WSRbool3, WSRbool3D);
    WSMFpropertyCreate(WSNanchorLeftFlag,WSCbool,_anchor_left_use,WSSanchorLeftFlag);
    WSMFpropertySetSelection(WSRbool3, WSRbool3D);
    WSMFpropertyCreate(WSNanchorRightFlag,WSCbool,_anchor_right_use,WSSanchorRightFlag);
    WSMFpropertySetSelection(WSRbool3, WSRbool3D);
#endif
#ifndef NO_REMOTE_INSTANCE
    WSMFpropertyCreate(WSNexport,WSCbool,_export_,WSSexport);
    WSMFpropertySetSelection(WSRbool1, WSRbool1D);
#endif
#ifndef NO_IMAGE
    WSMFpropertyCreate(WSNkeepPixmapCache,WSCbool,_keep_pixmap_cache,WSSkeepPixmapCache );
      WSMFpropertySetSelection(WSRbool4,WSRbool4D);
#endif
    WSMFaddTrigger(WSEV_INITIALIZE  );
    WSMFaddTrigger(WSEV_SENSITIVE_CH);
    WSMFaddTrigger(WSEV_PARENT_SENSITIVE_CH);
    WSMFaddTrigger(WSEV_MOUSE_MOVE  );
    WSMFaddTrigger(WSEV_MOUSE_IN    );
    WSMFaddTrigger(WSEV_MOUSE_OUT   );
    WSMFaddTrigger(WSEV_MOUSE_PRESS   );
    WSMFaddTrigger(WSEV_MOUSE_DOUBLE_CLICK );
    WSMFaddTrigger(WSEV_MOUSE_RELEASE );
    WSMFaddTrigger(WSEV_PARENT_VISIBLE_CH  );
    WSMFaddTrigger(WSEV_VISIBLE_CH  );
    WSMFaddTrigger(WSEV_EXPOSE      );
    WSMFaddTrigger(WSEV_RESIZE      );
  WSMFpropertyCreateEnd
}
WSMFproperty(WSCvform,WSNname, char*, _instance_name ,NULL);
WSMFproperty(WSCvform,WSNuserString, char*, _user_string ,NULL);
WSMFproperty(WSCvform,WSNuserValue, long,  _user_value, 0); 
WSMFproperty(WSCvform,WSNx,   short, _x, 100 );
WSMFproperty(WSCvform,WSNy,   short, _y, 100 );
WSMFproperty(WSCvform,WSNwidth, WSCushort, _w, 100  );
WSMFproperty(WSCvform,WSNheight, WSCushort, _h, 30  );
WSMFproperty(WSCvform,WSNshadowThickness, WSCuchar, _shadow_thick, 2);
WSMFproperty(WSCvform,WSNforeColor, short, _fore_color, WS_DF_FORECOLOR);
WSMFproperty(WSCvform,WSNbackColor, short, _back_color, WS_DF_BACKCOLOR);
WSMFproperty(WSCvform,WSNtopShadowColor, short, _ts_color, WS_DF_TOPSHADOWCOLOR);
WSMFproperty(WSCvform,WSNbottomShadowColor,short, _bs_color, WS_DF_BOTTOMSHADOWCOLOR);
#ifndef NO_IMAGE
WSMFproperty(WSCvform,WSNbackPixmap, short, _bp_color, -1);
#endif
WSMFproperty(WSCvform,WSNshadowType, char,    _shadow_type ,WS_SHADOW_OUT );
WSMFproperty(WSCvform,WSNvis,WSCbool, _vis ,False  );
WSMFproperty(WSCvform,WSNdet,WSCbool, _sensitive_work, True );
#ifndef NO_MOUSE_PROPERTY
WSMFproperty(WSCvform, WSNmouse,   WSCushort,     _mouse_no     ,0);
#endif
#ifndef NO_GRADATION
WSMFproperty(WSCvform, WSNgradation, WSCuchar, _grad_no   ,0);
WSMFproperty(WSCvform, WSNgradationMargin, WSCuchar, _grad_margin ,0);
#endif
#ifndef NO_ANCHOR
WSMFproperty( WSCvform, WSNanchorTop,WSCushort,_anchor_top,0 );
WSMFproperty( WSCvform, WSNanchorBottom,WSCushort,_anchor_bottom,0 );
WSMFproperty( WSCvform, WSNanchorLeft,WSCushort,_anchor_left,0 );
WSMFproperty( WSCvform, WSNanchorRight,WSCushort,_anchor_right,0 );
WSMFproperty( WSCvform, WSNanchorTopFlag,WSCbool,_anchor_top_use,False );
WSMFproperty( WSCvform, WSNanchorBottomFlag,WSCbool,_anchor_bottom_use,False );
WSMFproperty( WSCvform, WSNanchorLeftFlag,WSCbool,_anchor_left_use,False );
WSMFproperty( WSCvform, WSNanchorRightFlag,WSCbool,_anchor_right_use,False );
#endif
#ifndef NO_REMOTE_INSTANCE
WSMFproperty( WSCvform, WSNexport,WSCbool,_export_,False );
#endif
#ifndef NO_IMAGE
WSMFproperty( WSCvform, WSNkeepPixmapCache, WSCbool,  _keep_pixmap_cache,True );
#endif
void WSCvform::setWorkWSNname(char* ){}
void WSCvform::getWorkWSNname(char** ){}
void WSCvform::setWorkWSNuserString(char* ){}
void WSCvform::getWorkWSNuserString(char** ){}
void WSCvform::setWorkWSNuserValue(long ){}
void WSCvform::getWorkWSNuserValue(long* ){}
void WSCvform::setWorkWSNx(short data){
#ifndef NO_ANCHOR
  if (_anchor_left_use != False){
    WSCbase* parent = getParent();
    if (parent != NULL && parent->adjustAnchor() != False){
        _x = _anchor_left;
    }
  }else if (_anchor_right_use != False){
    WSCbase* parent = getParent();
    if (parent != NULL && parent->adjustAnchor() != False){
      parent->setGeometryChanged(True);
      parent->needUpdate();
      _geometry_changed = True;
      return;
    }
  }
#endif
  _geometry_changed = True;
  WSDdev* dev = getowndev();
  if (dev != NULL){
    dev->setValue(WSDEV_X,&_x);
  }
  return;
}
void WSCvform::getWorkWSNx(short* ){}
void WSCvform::setWorkWSNy(short data){
#ifndef NO_ANCHOR
  if (_anchor_top_use != False){
    WSCbase* parent = getParent();
    if (parent != NULL && parent->adjustAnchor() != False){
      _y = _anchor_top;
    }
  }else if (_anchor_bottom_use != False){
    WSCbase* parent = getParent();
    if (parent != NULL && parent->adjustAnchor() != False){
      parent->setGeometryChanged(True);
      parent->needUpdate();
      _geometry_changed = True;
      return;
    }
  }
#endif
  _geometry_changed = True;
  WSDdev* dev = getowndev();
  if (dev != NULL){
    dev->setValue(WSDEV_Y,&_y);
  }
}

void WSCvform::getWorkWSNy(short* ){}
void WSCvform::setWorkWSNwidth(WSCushort data){
  if (_initialized != False && _dev != NULL){
#ifndef NO_ANCHOR
    WSCbase* parent = getParent();
    if (parent != NULL && parent->adjustAnchor() != False){
      if (_anchor_left_use != False && _anchor_right_use != False){
        parent->setGeometryChanged(True);
        parent->needUpdate();
        return;
      }
    }
#endif
    _dev->setValue(WSDEV_WIDTH,(void*)&data);
    _geometry_changed = True;
    _update_resize = True;
#ifndef NO_GRADATION
    if (_grad_no != 0){
      setAbsoluteDraw(True);
      needUpdate();
    }
#endif
  }else{
    _geometry_changed = True;
    _update_resize = True;
  }
}
void WSCvform::getWorkWSNwidth(WSCushort* ){}
void WSCvform::setWorkWSNheight(WSCushort data){
  if (_initialized != False && _dev != NULL){
#ifndef NO_ANCHOR
    WSCbase* parent = getParent();
    if (parent != NULL && parent->adjustAnchor() != False){
      if (_anchor_top_use != False && _anchor_bottom_use != False){
        parent->setGeometryChanged(True);
        parent->needUpdate();
        return;
      }
    }
#endif
    _dev->setValue(WSDEV_HEIGHT,(void*)&data);
    _geometry_changed = True;
    _update_resize = True;
#ifndef NO_GRADATION
    if (_grad_no != 0){
      setAbsoluteDraw(True);
      needUpdate();
    }
#endif
  }else{
    _geometry_changed = True;
    _update_resize = True;
  }
}
void WSCvform::getWorkWSNheight(WSCushort* ){}
void WSCvform::setWorkWSNshadowThickness(WSCuchar ){}
void WSCvform::getWorkWSNshadowThickness(WSCuchar* ){}
void WSCvform::setWorkWSNforeColor(short ){}
void WSCvform::getWorkWSNforeColor(short* ){}
void WSCvform::setWorkWSNbackColor(short data){
  if (_initialized != False && _dev != NULL){
    _dev->setValue(WSDEV_BACKCOLOR,(void*)&data);
  }
  _a1_color = -1;
  _a2_color = -1;
  _a3_color = -1;
}
void WSCvform::getWorkWSNbackColor(short* ){}
void WSCvform::setWorkWSNtopShadowColor(short ){}
void WSCvform::getWorkWSNtopShadowColor(short* ){}
void WSCvform::setWorkWSNbottomShadowColor(short ){}
void WSCvform::getWorkWSNbottomShadowColor(short* ){}
#ifndef NO_IMAGE
void WSCvform::setWorkWSNbackPixmap(short data){
  if (_initialized != False && _dev != NULL){
    _dev->setValue(WSDEV_BACK_PIXMAP,(void*)&data);
  }
}
void WSCvform::getWorkWSNbackPixmap(short* data){
  if (_initialized != False && _dev != NULL){
    _dev->setValue(WSDEV_BACK_PIXMAP,(void*)data);
  }
}
#endif
void WSCvform::setWorkWSNshadowType(char ){}
void WSCvform::getWorkWSNshadowType(char* ){}
void WSCvform::setWorkWSNvis(WSCbool data){
    setVisible(data);
}
void WSCvform::getWorkWSNvis(WSCbool*){}
void WSCvform::setWorkWSNdet(WSCbool data){
#ifndef WS_EMBED
  if (getPropertyEditMode() == False){
#endif
    setSensitive(data);
#ifndef WS_EMBED
  }
#endif
}
void WSCvform::getWorkWSNdet(WSCbool* data){
}
void WSCvform::setNoRefresh(WSCbool fl){
  _no_refresh = fl;
}

#ifndef NO_MOUSE_PROPERTY
void WSCvform::setWorkWSNmouse(WSCushort data ){
  if (_initialized != False && _dev != NULL){
    _dev->setValue(WSDEV_MOUSE_NO,(void*)&_mouse_no);
  }
}
void WSCvform::getWorkWSNmouse(WSCushort* ){}
#endif
#ifndef NO_GRADATION
void WSCvform::setWorkWSNgradation(WSCuchar ){}
void WSCvform::getWorkWSNgradation(WSCuchar* ){}
void WSCvform::setWorkWSNgradationMargin(WSCuchar ){}
void WSCvform::getWorkWSNgradationMargin(WSCuchar* ){}
#endif

#ifndef NO_ANCHOR
void WSCvform::setWorkWSNanchorTop(WSCushort data){
  if (_anchor_top_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorTop(WSCushort* data){}
void WSCvform::setWorkWSNanchorBottom(WSCushort data){
  if (_anchor_bottom_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorBottom(WSCushort* data){}
void WSCvform::setWorkWSNanchorLeft(WSCushort data){
  if (_anchor_left_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorLeft(WSCushort* data){}
void WSCvform::setWorkWSNanchorRight(WSCushort data){
  if (_anchor_left_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorRight(WSCushort* data){}
void WSCvform::setWorkWSNanchorTopFlag(WSCbool data){
  if (_anchor_top_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorTopFlag(WSCbool* data){}
void WSCvform::setWorkWSNanchorBottomFlag(WSCbool data){
  if (_anchor_bottom_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorBottomFlag(WSCbool* data){}
void WSCvform::setWorkWSNanchorLeftFlag(WSCbool data){
  if (_anchor_left_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorLeftFlag(WSCbool* data){}
void WSCvform::setWorkWSNanchorRightFlag(WSCbool data){
  if (_anchor_right_use != False){
    if (getParent() != NULL){
      getParent()->setGeometryChanged(True);
      getParent()->needUpdate();
    }
  }
}
void WSCvform::getWorkWSNanchorRightFlag(WSCbool* data){}
#endif

#ifndef NO_REMOTE_INSTANCE
void WSCvform::setWorkWSNexport(WSCbool data){
  setExported(data);
}
void WSCvform::getWorkWSNexport(WSCbool* data){
}
#endif
#ifndef NO_IMAGE
void WSCvform::setWorkWSNkeepPixmapCache(WSCbool){}
void WSCvform::getWorkWSNkeepPixmapCache(WSCbool*){}
#endif
WSCbool WSCvform::getNoRefresh(){
  return _no_refresh;
}
void WSCvform::setChildren(WSClistData& src){
  _children = src;
}
WSClistData& WSCvform::getChildren(){
  return _children;
}
WSClistData* WSCvform::getChildrenPtr(){
  return &_children;
}
void WSCvform::onVisibleChange(WSCbool fl){
  long need_resize_event = 0;
  if (fl != False && getowndev() == NULL){
    need_resize_event = 1;
  }

  if (fl != False && _vis != False){
    if (getowndev() == NULL){
      WSDdev* dev = NULL;
      dev = WSGIappDevice()->getDevObject("nwDev");
      setdev(dev);
      dev->setEnableEventBit(WSEV_EXPOSE_BIT |  WSEV_MOUSE_MOVE_BIT | WSEV_MOUSE_IN_BIT | WSEV_MOUSE_OUT_BIT);
    }
  }

  WSCbase::onVisibleChange(fl);
  if (need_resize_event != 0){
    WSCrect rect;
    rect.x = _x;
    rect.y = _y;
    rect.width = _w;
    rect.height = _h;
    execEventProc(WSEV_RESIZE,(void*)&rect);
  }
}
void WSCvform::onParentVisibleChange(WSCbool fl){
  long need_resize_event = 0;
  if (fl != False && getowndev() == NULL){
    need_resize_event = 1;
  }

  if (fl != False && _vis != False){
    if (getowndev() == NULL){
      WSDdev* dev = NULL;
      dev = WSGIappDevice()->getDevObject("nwDev");
      setdev(dev);
    }
  }
  WSDdev* dev = getowndev();
  if (dev != NULL){
    WSCulong event_bit = dev->getEventMaskBit();
    if (!(event_bit & WSEV_EXPOSE_BIT)){
      dev->setEnableEventBit(event_bit | WSEV_EXPOSE_BIT);
    }
  }

  WSCbase::onParentVisibleChange(fl);
  if (need_resize_event != 0){
    WSCrect rect;
    rect.x = _x;
    rect.y = _y;
    rect.width = _w;
    rect.height = _h;
    execEventProc(WSEV_RESIZE,(void*)&rect);
  }
}
void WSCvform::onResize(WSCrect* area){
  WSCbool fl = False;
#ifndef WS_EMBED
  if (getPropertyEditMode() == False){
#endif
    _x = area->x;
    _y = area->y;
    if (_w != area->width){
      _w = area->width;
      fl = True;
    }
    if (_h != area->height){
      _h = area->height;
      fl = True;
    }
#ifndef WS_EMBED
  }
#endif
  if (fl != False){
    setAbsoluteDraw(True);
    needUpdate();
  }
}
WSCvform::~WSCvform(){
  WSGIappObjectList()->delEvent(this,WSEV_GUI_POLICY_CH);

  WSClistData* ghlist = (WSClistData*)getUserData("GH-LIST");
  if (ghlist != NULL){
    delete ghlist;
    setUserData("GH-LIST",(void*)0);
  }
  while(1){
    if (_children.getNum() == 0){
      break;
    }
    WSCbase* child = (WSCbase*)_children[0];
    delete child;
  }
  if (_user_string != NULL){
    delete[] _user_string;
    _user_string = NULL;
  }
}
long WSCvform::_device_initialize(){
  WSDdev* dev = getowndev();
  if (dev == NULL){
    return WS_ERR;
  }
  dev->setVisible(False);
//#ifndef NO_SCALE_OFFSET
//  if (getXOffsetPtr() == NULL){
//#endif
    dev->setValue(WSDEV_X,(void*)&_x);
//#ifndef NO_SCALE_OFFSET
//  }else{
//    short ox = _x + *(getXOffsetPtr());
//    dev->setValue(WSDEV_X,(void*)&ox);
//  }
//  if (getYOffsetPtr() == NULL){
//#endif
    dev->setValue(WSDEV_Y,(void*)&_y);
//#ifndef NO_SCALE_OFFSET
//  }else{
//    short oy = _y + *(getYOffsetPtr());
//    dev->setValue(WSDEV_Y,(void*)&oy);
//  }
//#endif
#ifndef NO_MOUSE_PROPERTY
  dev->setValue(WSDEV_MOUSE_NO,(void*)&_mouse_no);
#endif
  dev->setValue(WSDEV_WIDTH,(void*)&_w);
  dev->setValue(WSDEV_HEIGHT,(void*)&_h);
  dev->setValue(WSDEV_BACKCOLOR,(void*)&_back_color);
  dev->setValue(WSDEV_BACK_PIXMAP,(void*)&_bp_color);
  if (getVisible() != False){
    dev->setVisible(True);
  }
  WSCbase::_device_initialize();
  dev->setEnableEvent(WSEV_EXPOSE);
  return WS_NO_ERR;
}
long WSCvform::getObjectType(){
  return (WS_TYPE_MANAGER | WS_TYPE_FORM | WS_TYPE_NON_WINDOW);
}
long WSCvform::draw(){
dbprintf("WSCvform::draw %s:%d %s begin..\n",__FILE__,__LINE__,getInstanceName());
  if (_no_refresh != False){
    WSCbase::update();
    setAbsoluteDraw(False);
dbprintf("WSCvform::draw %s:%d %s no refresh..\n",__FILE__,__LINE__,getInstanceName());
    return WS_NO_ERR;
  }
  if (getVisible() == False){
dbprintf("WSCvform::draw %s:%d %s vis=False..\n",__FILE__,__LINE__,getInstanceName());
    return WS_NO_ERR;
  }
  WSDdev* dev = getowndev();
  if (dev == NULL){
dbprintf("WSCvform::draw %s:%d %s dev=NULL..\n",__FILE__,__LINE__,getInstanceName());
    return WS_ERR;
  }
  short x = _x;
  short y = _y;
  WSCbool absolute = getAbsoluteDraw(); 
  if (absolute == False  ){
    if ( dev->isExposed(x,y,_w,_h) == False ){
dbprintf("WSCvform::draw %s:%d %s not exposed..\n",__FILE__,__LINE__,getInstanceName());
      return WS_NO_ERR;
    }
  }
  if (_update_resize != False){
    _update_resize = False;
    execResizeEvent();
  }

#ifndef NO_ANCHOR
  if (_geometry_changed != False){
dbprintf("WSCvform::draw %s:%d %s call _exec_child_adjust_for_anchors()..\n",__FILE__,__LINE__,getInstanceName());
    _exec_child_adjust_for_anchors();
  }
#endif
#ifndef NO_GUI_POLICY
  if (WSGIappDev()->getExtGuiPolicySelected() != False){
    if (_ext_draw_proc == NULL){
      _ext_draw_proc = (void(*)(void*))WSGIappDev()->
                               getExtGuiPolicyProc("WSCvform::draw");
    }
    if (_no_ext_gui_proc != False){
      _ext_draw_proc = NULL;
    }
    if (_ext_draw_proc != NULL){
      WSCvform_draw_struct st;
      st._this = this;
      st.dev = dev;
      st.x = _x;
      st.y = _y;
      st.width = _w;
      st.height = _h;
      st.shadow_thickness = _shadow_thick;
      st.shadow_type = _shadow_type;
      st.fore_color = _fore_color;
      st.back_color = _back_color;
      st.top_shadow_color = _ts_color;
      st.bottom_shadow_color = _bs_color;
      st.back_pixmap = _bp_color;
#ifndef NO_GRADATION
      st.gradation = _grad_no;
      st.gradation_margin = _grad_margin;
#endif
      _ext_draw_proc(&st);
      return WS_NO_ERR;
    }  
  }  
#endif
  long err = dev->beginDraw(x,y,_w,_h,absolute); 
  if (err != WS_NO_ERR){
dbprintf("WSCvform::draw %s:%d %s dev not ready..\n",__FILE__,__LINE__,getInstanceName());
    return WS_NO_ERR;
  }
  short ex = 0;
  short ey = 0;
  WSCushort ew = _w;
  WSCushort eh = _h;
  if (absolute == False){
    dev->getExposedArea(&ex,&ey,&ew,&eh);
    ex -= _x;
    ey -= _y;
  }
  WSCbase::update();
  setAbsoluteDraw(False);

  if (_bp_color == -1){
#ifndef NO_GRADATION
    if (_grad_no == 0){
#endif
      dev->setForeColor(_back_color);
      dev->drawFillRect(0,0,_w,_h);
      dev->setForeColor(_back_color);

#ifndef NO_GRADATION
    }else{
      dev->drawGradation(_grad_no,_ts_color,_back_color,_bs_color,0,0,_w,_h,_grad_margin);
    }
#endif
  }

  if (getDotMode() == True){
    dev->setForeColor(_fore_color);
    long i,j;
    for(i=0; i<_w; i+=10){
      for(j=0; j<_h; j+=10){
        if (ex < i+1 && i < ex+ ew+1 &&
            ey < j+1 && j < ey+ eh+1 ){
          dev->drawFillRect(i,j,1,1);
        }
      }
    }
  }
  WSGFdrawShadow(dev,_shadow_type,_shadow_thick,_fore_color,_ts_color,_bs_color,_back_color,0,0,_w,_h);
  dev->endDraw();
  return WS_NO_ERR;
}
#ifndef NO_ANCHOR
void WSCvform::_exec_child_adjust_for_anchors(){
  if (adjustAnchor() != False){
    if (_geometry_changed != False){
      WSClistData children = getChildren();
      long num = children.getNum();
      long i;
      WSCbase** buf = (WSCbase**)children._data;
      for(i=0; i<num; i++){
        WSCbase* child = buf[i];
        child->_adjust_for_anchors(_w,_h);
      } 
    }
  }
}
#endif

void WSCvform::update(){
  if (_geometry_changed == True){
    WSDdev* dev = getowndev();
    if (dev != NULL){
      dev->setValue(WSDEV_X,&_x);
      dev->setValue(WSDEV_Y,&_y);
    }
  }
  if (isNeedUpdate() == True){
    redraw();
    WSCbase::update();
  }
}
long WSCvform::clear(){
  WSDdev* dev = getowndev();
  if (dev == NULL){
    return WS_ERR;
  }
  dev->clearArea(_x, _y, _w, _h, False);
  return WS_NO_ERR;
}
long WSCvform::clearPixmap(){
  WSDdev* dev = getowndev();
  if (dev == NULL){
    return WS_ERR;
  }
  dev->erase(_x, _y, _w, _h, False);
  return WS_NO_ERR;
}
long WSCvform::redraw(){
  WSDdev* dev = getowndev();
  if (dev == NULL){
    return WS_ERR;
  }
  WSDdev* pdev = dev->getEventParentDev();
  if (pdev != NULL){
    WSCbase* parent = pdev->getAttachedClient();
    if (parent != NULL && parent->getAbsoluteDraw() != False){
      parent->setAbsoluteDraw(False);
      parent->redraw();
      return WS_NO_ERR;
    }
  }
  setAbsoluteDraw(False);
  dev->clearArea(_x,_y,_w,_h,True);
  return WS_NO_ERR;
}

long WSCvform::cdraw(){
  WSDdev* dev = getowndev();
  if (dev == NULL){
    return WS_ERR;
  }
  if (getVisible() == False){
    return WS_ERR;
  }
  WSCrect area; 
  area.x = _x;
  area.y = _y;
  area.width = _w;
  area.height = _h;

  short bx,by;
  WSCushort bw,bh; 
  dev->getExposedArea(&bx,&by,&bw,&bh);
  dev->setExposedArea(_x,_y,_w,_h);
  dev->execEvent(WSEV_EXPOSE,&area);
  dev->setExposedArea(bx,by,bw,bh);
  return WS_NO_ERR;
}
void WSCvform::execResizeEvent(){
  WSCrect rect;
  rect.x = _x;
  rect.y = _y;
  rect.width = _w;
  rect.height = _h;
  execEventProc(WSEV_RESIZE,(void*)&rect);
}
#ifndef NO_ANCHOR
void WSCvform::_adjust_for_anchors(WSCushort w,WSCushort h){
  long resized = False;
  if (w > 0){
    if (_anchor_left_use == 0){
      if (_anchor_right_use != 0){
        long nx = w - _anchor_right - _w;
        _anchor_right_use = False;
        setProperty(WSNx,nx);
        _anchor_right_use = True;
      }
    }else{
      if (_anchor_right_use != 0){
        _anchor_left_use = False;
        _anchor_right_use = False;
        if (w < _anchor_right + _anchor_left){
          setProperty(WSNwidth,1);
          resized = True;
        }else{
          setProperty(WSNx,_anchor_left);
          setProperty(WSNwidth,w - _anchor_right - _anchor_left);
          resized = True;
        }
        _anchor_left_use = True;
        _anchor_right_use = True;
      }else{
        if (_anchor_left != _x){
          setProperty(WSNx,_anchor_left);
        }
      }
    }
    needUpdate();
    setAbsoluteDraw(True);
  }
  if (h > 0){
    if (_anchor_top_use == 0){
      if (_anchor_bottom_use != 0){
        long ny = h - _anchor_bottom - _h;
        _anchor_bottom_use = False;
        setProperty(WSNy,ny);
        _anchor_bottom_use = True;
      }
    }else{
      if (_anchor_bottom_use != 0){
        _anchor_top_use = False;
        _anchor_bottom_use = False;
        if (h < _anchor_bottom + _anchor_top){
          setProperty(WSNheight,1);
          resized = True;
        }else{
          setProperty(WSNy,_anchor_top);
          setProperty(WSNheight,h - _anchor_top - _anchor_bottom);
          resized = True;
        }
        _anchor_top_use = True;
        _anchor_bottom_use = True;
      }else{
        if (_anchor_top != _y){
          setProperty(WSNy,_anchor_top);
        }
      }
    }
    needUpdate();
    setAbsoluteDraw(True);
  }
  if (resized != False){
    WSCrect rect;
    rect.x = _x;
    rect.y = _y;
    rect.width = _w;
    rect.height = _h;
    execEventProc(WSEV_RESIZE,&rect);
  }
}
#endif
long WSCvform::execEventProc(long ev, void* data){
  if (ev == WSEV_EXPOSE){
    if (_x + _w < 0 || _y + _h < 0){
      return WS_NO_ERR;
    }
    WSCbool absolute = getAbsoluteDraw();
    long ret = draw();
    if (ret == WS_ERR){
      return WS_ERR;
    }
    WSDdev* dev = getowndev();
    if (dev == NULL){
      return WS_ERR;
    }
    if (absolute == False && dev->isExposed(_x,_y,_w,_h) == False){
      return WS_NO_ERR;
    }
    if (data == NULL){
      WSCrect area;
      area.setRect(_x,_y,_w,_h);
      onExpose(&area);
      if (_expose_op == 1){
        execProcedure(ev);
      }
      return WS_NO_ERR;
    }else{
      WSCrect* area = (WSCrect*)data;
      onExpose(area);
      if (_expose_op == 1){
        execProcedure(ev);
      }
    }
    return WS_NO_ERR;
  }
  if (ev == WSEV_MOUSE_PRESS  || ev == WSEV_MOUSE_RELEASE ||
      ev == WSEV_MOUSE_DOUBLE_CLICK ||
      ev == WSEV_MOUSE_IN     || ev == WSEV_MOUSE_OUT    ||
      ev == WSEV_MOUSE_MOVE ){
    if (ev == WSEV_MOUSE_OUT){
      if (_mouse_enter == True){
        _mouse_enter = False;
        WSCbase::execEventProc(WSEV_MOUSE_OUT, NULL);
      }
      return WS_NO_ERR;
    }else{
      if (data == NULL){
        return WS_ERR;
      }
      WSCpoint* pt = (WSCpoint*)data;
      if ( pt->x >= _x && pt->x < _x+_w &&
           pt->y >= _y && pt->y < _y+_h ){
        WSCpoint vpdata;
        vpdata.x = pt->x - _x;
        vpdata.y = pt->y - _y;
        if (ev == WSEV_MOUSE_PRESS){
          _mouse_grabed = True;
          onGetMouseFocus();
        }else if (ev == WSEV_MOUSE_RELEASE){
          _mouse_grabed = False;
        }
        WSCbase::execEventProc(ev, &vpdata);
        if (_mouse_enter == False){
          _mouse_enter = True;
          WSCbase::execEventProc(WSEV_MOUSE_IN, &vpdata);
        }
        return WS_ERR;
      }else{
        WSCpoint vpdata;
        vpdata.x = pt->x - _x;
        vpdata.y = pt->y - _y;
        if (_out_side_mouse_move == True &&  ev == WSEV_MOUSE_MOVE  ){
          if (_mouse_grabed != False){
            WSCbase::execEventProc(ev, &vpdata);
            return WS_ERR;
          }
        }
        if (_out_side_mouse_release == True &&  ev == WSEV_MOUSE_RELEASE ){
          if (_mouse_grabed != False){
            _mouse_grabed = False;
            WSCbase::execEventProc(ev, &vpdata);
            return WS_NO_ERR;
          }
        }
        if (_out_side_mouse_press == True &&  ev == WSEV_MOUSE_PRESS ){
          if (_mouse_grabed != False){
            WSCbase::execEventProc(ev, &vpdata);
            return WS_NO_ERR;
          }
        }
        if (_mouse_enter == True){
          _mouse_enter = False;
          WSCbase::execEventProc(WSEV_MOUSE_OUT, NULL);
        }
        return WS_ERR;
      }
      return WS_ERR;
    }
  }
  return WSCbase::execEventProc(ev,data);
}

void(*WSCvform::_ext_draw_proc)(void*) = NULL;
char* WSCvform::_policy_name = NULL;
#ifndef NO_GUI_POLICY
void WSCvform::onGuiPolicyChange(long policy){
  WSCbase::onGuiPolicyChange(policy);
  if (WSGIappDev()->getExtGuiPolicySelected() == False){
    _ext_draw_proc = NULL;
  }else{
    if (_policy_name != NULL &&
        !strcmp(_policy_name,(char*)WSGIappDev()->getExtGuiPolicyName())){
    }else{
      _ext_draw_proc = (void(*)(void*))WSGIappDev()->getExtGuiPolicyProc("WSCvform::draw");
      if (_policy_name != NULL){
        delete _policy_name;
      }
      _policy_name = WSGFstrdup(WSGIappDev()->getExtGuiPolicyName());
    }
  }
  needUpdate();
  setAbsoluteDraw(True);
  redraw();
}
#endif
void WSCvform::internalUpdate(){
#if 0
  WSDdev* dev = getowndev();
  if (dev != NULL){
    char pstyle = WS_DYNAMIC_PIXMAP;
    dev->getValue(WSDEV_PIXMAP_STYLE,(void*)&pstyle);
    if (pstyle == WS_DYNAMIC_PIXMAP){
      WSDwindowDev* wdev = (WSDwindowDev*)dev->cast("WSDwindowDev");
      wdev->destroyPixmap();
    }
  }
#endif
}
#ifdef WS_EMBED
#ifndef NO_EXT_GUI_POLICY
class _gui_policy_register_{
  public: _gui_policy_register_(){
extern void WSCvform_draw_mdr1(void*);
    WSGIappGuiPolicy()->setExtGuiPolicyProc("Modern1","WSCvform::draw",WSCvform_draw_mdr1);
  }
};
static _gui_policy_register_ _run_register;
#endif
#endif

long WSCvform::expose(){
  if (getVisible() == False){
dbprintf("WSCvform::draw %s:%d %s vis=False\n",__FILE__,__LINE__,getInstanceName());
    return WS_NO_ERR;
  }
  WSDdev* dev = getowndev();
  if (dev == NULL){
dbprintf("WSCvform::draw %s:%d %s dev=NULL\n",__FILE__,__LINE__,getInstanceName());
    return WS_ERR;
  }
  WSCrect area;
  area.setRect(_x,_y,_w,_h);
  dev->onExpose(&area);
  return WS_NO_ERR;
}
void WSCvform::setNoExtGuiProc(WSCbool fl){
  _no_ext_gui_proc = fl;
}
void WSCvform::onOrderChanged(char direction){
  WSClistData children = getChildren();
  int num = children.getNum();
  int i;
  if (direction == WS_TOP){
    for(i=num-1; i>-1; i--){
      WSCbase* child = (WSCbase*)children[i];
      child->setOrder(direction);
    }
    WSCbase::onOrderChanged(direction);
  }else{
    WSCbase::onOrderChanged(direction);
    for(i=0; i<num; i++){
      WSCbase* child = (WSCbase*)children[i];
      child->setOrder(direction);
    }
  }
}
void WSCvform::onParentChanged(WSCbase* parent){
  if (getVisible() != False){
    clear();
  }
  WSCbase::onParentChanged(parent);
}

