//
// Copyright (C) 1999-2008 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#ifndef NO_EXECUTE
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/wait.h>

#include <devfb/devfb.h>
#include <devfb/devfb_dep.h>

#include <WScom.h>
#include <WSDfile.h>
#include <devfb/WSDdevfbExecute.h>
#include <WSCbaseList.h>
#include <WSDfileSystem.h>
#include <WSClocaleSet.h>

//#define PDEBUG

WSMFclassInit(WSDdevfbExecute,WSDexecute);

WSDexecute* _unix_execute_chandler(){
  return new WSDdevfbExecute;
}
void WSGFdevfbExecuteInit(){
    WSDexecute::setCreateInstanceHandler((void*)_unix_execute_chandler);
}
#ifndef NO_GLOBAL_CONSTRUCTORS
class _unix_execute_init {
  public: _unix_execute_init(){
    WSGFdevfbExecuteInit();
  };
};
static _unix_execute_init  _unix_execute_init_execute;
#endif

WSDdevfbExecute::WSDdevfbExecute(){
//  _xid = 0;
  _handle = (void*)NULL;
  _id = 0;
  _sem = cre_sem(1,SEM_SYNC|DELEXIT );
}
WSDdevfbExecute::~WSDdevfbExecute(){
#ifdef PDEBUG
  printf("WSDdevfbExecute::~WSDdevfbExecute started.\n");
  dbprintf("WSDdevfbExecute::~WSDdevfbExecute started.\n");
#endif
  if (_pid != 0){
    killProcess();
  }

  if (_handle != NULL){
    pclose((FILE*)_handle);
    return;
  }
  if (_sem != 0){
//printf("WSDdevfbExecute::~WSDdevfbExecute() _sem = NULL\n");
    del_sem(_sem);
    _sem = 0;
  }
#ifdef PDEBUG
  printf("WSDdevfbExecute::~WSDdevfbExecute done.\n");
  dbprintf("WSDdevfbExecute::~WSDdevfbExecute done. \n");
#endif
}
void WSDdevfbExecute::_callback(WSDdevfbExecute* obj){
//printf("WSDdevfbExecute::_callback called\n");
  if (obj->_data_receive != NULL){
    obj->_data_receive(obj);
//printf("WSDdevfbExecute::_callback unlock _sem = 0x%x\n",sem);
    WSGIappObjectList()->execUpdate();
    return;
  }
//printf("WSDdevfbExecute::_callback _sem = 0x%x\n",sem);
}

#define READ_PORT 0
#define WRITE_PORT 1

long WSDdevfbExecute::execute(char* exename,long mode){
  killProcess();


  wai_sem(_sem,T_FOREVER);
  _string = "";
  _pos = 0; 
  sig_sem(_sem);

  if (exename == NULL){
    return WS_ERR;
  }
#ifdef PDEBUG
  printf("WSDdevfbExecute::execute #%s#\n",exename);
  dbprintf("WSDdevfbExecute::execute #%s#\n",exename);
#endif
  _exe_name = WSGIappFileSystem()->adjustFileName(exename);
  _exe_name.delHeadSpace();
  char* exe_name = _exe_name.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  char* dir_name = _dir_name.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());

  if (strcmp((char*)dir_name,"")){
    int ret = chdir((char*)dir_name);
    if (ret != 0){
#ifdef PDEBUG
  printf("WSDdevfbExecute::execute chdir err.\n");
  dbprintf("WSDdevfbExecute::execute chdir err.\n");
#endif
      return WS_ERR;
    }
  }
  if (mode == WS_EXEC_RDWR || mode == WS_EXEC_RD){
    if (mode == WS_EXEC_RD){
      _handle = popen(exe_name,"r");
    }else if (mode == WS_EXEC_WR){
      _handle = popen(exe_name,"w");
    }else{
      _handle = popen(exe_name,"rw");
    }
    if (_handle == NULL){
      return WS_ERR;
    }
    if (mode != WS_EXEC_WR && _data_receive){
      int ret = cre_tsk(_read_thread_,-1,(int)this);
      if (ret < 0){
        return WS_ERR;
      }
    }
    _mode = mode;
    return WS_NO_ERR;
  }

  long num = _exe_name.getWords();
  if (num == 1){
    _pid  = (int)fork();
    if (_pid == 0){
      int ret = execl((char*)exe_name,exename,NULL);
      if (ret == -1){
        WSGFexit(127);
      }else{
        WSGFexit(0);
      }
    }
  }else if (num > 1){
//printf("file=#%s#\n",(char*)_exe_name);
    WSCstring t1(exe_name);
    char** args = new char*[num+1];
    long i;
    for(i=0; i<num; i++){
      args[i] = WSGFstrdup((char*)t1.getWord(i));
//printf("args[%d]=%s\n",i,args[i]);
    }
    args[num] = NULL;
    _pid  = (int)fork();
    if (_pid == 0){
      int ret = execv(args[0],args);
      if (ret == -1){
        WSGFexit(127);
      }else{
        WSGFexit(0);
      }
    }
    for(i=0; i<num; i++){
      delete args[i];
    }
    delete args;
    int status;
    waitpid(_pid,&status,WNOHANG);
    if (WEXITSTATUS(status) == 127){
      return WS_ERR;
    }
  }

  return WS_NO_ERR;
}
long WSDdevfbExecute::killProcess(){
#ifdef PDEBUG
  printf("WSDdevfbExecute::killProcess started..\n");
  dbprintf("WSDdevfbExecute::killProcess started..\n");
#endif
  if (_handle != (void*)NULL){
    pclose((FILE*)_handle);
    _handle = (void*)NULL;
  }
  if (_pid == 0){
    _id++;
    return WS_NO_ERR;
  }
//TEST
//  kill(_pid,SIGKILL);
  wai_sem(_sem,T_FOREVER);
  kill(_pid,SIGKILL);

  while(1){
    int dust;
    if ( (int)waitpid(_pid,&dust,0) < 0){
      if (errno == ECHILD){
        break;
      }
      continue;
    }
    break;
  }

  sig_sem(_sem);
  _pid = 0;
  _id++;
#ifdef PDEBUG
  printf("WSDdevfbExecute::killProcess done.\n");
  dbprintf("WSDdevfbExecute::killProcess done.\n");
#endif

  return WS_NO_ERR;
}

long WSDdevfbExecute::read(char* buf,long len){
//printf("WSDdevfbExecute::read this=0x%x\n",this);
  if (len < 1){
    return 0;
  }
  if (_handle == 0){
    return -1;
  }
  if (_mode == WS_EXEC_RD || _mode == WS_EXEC_RDWR){
    wai_sem(_sem,T_FOREVER);
    char* str = _string.getString();
    long slen1 = strlen(str);
    long slen2 = _pos + len;
    if (slen1 < slen2){
      len = slen1 - _pos;
    }
    if (len < 1){
      return 0;
    }
    memcpy(buf,&str[_pos],len);
    _pos += len;
    sig_sem(_sem);

    return len;
  }
  return -1;
}
char* WSDdevfbExecute::gets(char* buf,long len){
  if (len < 1){
    return 0;
  }
  if (_pid == 0){
#ifdef PDEBUG
  printf("WSDdevfbExecute::gets here. this=0x%x return=0\n",this);
  dbprintf("WSDdevfbExecute::gets here. this=0x%x return=0\n",this);
#endif
    return 0;
  }
  if (_mode == WS_EXEC_RD || _mode == WS_EXEC_RDWR){
    wai_sem(_sem,T_FOREVER);
    char* str = _string.getString();
    long slen1 = strlen(str);
    long slen2 = _pos + len;
    if (slen1 < slen2){
      len = slen1 - _pos;
    }
    if (len < 1){
      return 0;
    }
    int i=0;
    for(i=0; i< len-2; i++){
      buf[i] = str[_pos + i];
      if (buf[i] == '\n'){
        break;
      }
    }
    buf[i] = '\n';
    i++;
    buf[i] = 0;
    _pos += len;
    sig_sem(_sem);
#ifdef PDEBUG
  printf("WSDdevfbExecute::gets #%s#..\n",buf);
  dbprintf("WSDdevfbExecute::gets #%s#..\n",buf);
#endif
    return buf;
  }
#ifdef PDEBUG
  printf("WSDdevfbExecute::gets null2..\n");
  dbprintf("WSDdevfbExecute::gets null2..\n");
#endif
  return 0;
}


long WSDdevfbExecute::write(char* buf,long len){
  if (_pid == 0){
    return -1;
  }
  if (_mode == WS_EXEC_WR || _mode == WS_EXEC_RDWR){
    if (_handle != (void*)0){
      return fwrite(buf,1,len,(FILE*)_handle);
    }
  }
  return -1;
}
void WSDdevfbExecute::_watch_pid_(int ptr){
#ifdef PDEBUG
  printf("WSDdevfbExecute::_watch_pid_.. start tid=%d\n",get_tid());
  dbprintf("WSDdevfbExecute::_watch_pid_.. start tid=%d\n",get_tid());
#endif
  WSDdevfbExecute* _this = (WSDdevfbExecute*)ptr;
  pid_t pid = _this->_pid;
  long id = _this->_id;
  if (pid > 0){
    int status;
    while(1){
      pid_t ret = waitpid(pid,&status,WNOHANG);
      if (ret == 0){
#ifdef PDEBUG
  printf("status=0x%x ret=%d eno=%d\n",status,ret,errno);
  dbprintf("status=0x%x ret=%d eno=%d\n",status,ret,errno);
#endif
        wai_prc(100);
        continue;
      }
      break;
    }
    if (pid == _this->_pid && id == _this->_id){
      _this->_pid = 0;
    }
  }

#ifdef PDEBUG
  printf("WSDdevfbExecute::_watch_pid_ close handles..\n");
  dbprintf("WSDdevfbExecute::watch_pid_ _close handles..\n");
#endif
  FILE* h =(FILE*)_this->_handle;
  if (h != 0){
    pclose(h);
    _this->_handle = (void*)0;
  }

  wai_sem(_this->_sem,T_FOREVER);
#ifdef PDEBUG
  printf("WSDdevfbExecute::_watch_pid_ close handles..\n");
  dbprintf("WSDdevfbExecute::_watch_pid_ close handles..\n");
  dbprintf("WSDdevfbExecute::_watch_pid_ tsk1=%d tsk2=%d\n",_this->_tsk2,_this->_tsk3);
#endif

  sig_sem(_this->_sem);

  _read_event_send_type data;
  data.obj = _this;
  data.id = _this->_id;

  MESSAGE* msg;
  char buffer[128];
  msg = (MESSAGE*)buffer;
  msg->msg_type = MS_TYPE3;
  msg->msg_size = sizeof(data);
  memcpy(msg->msg_body.ANYMSG.msg_str,&data,sizeof(data));

#ifdef PDEBUG
  printf("WSDdevfbExecute::_watch_pid_ pid exited..\n");
  dbprintf("WSDdevfbExecute::_watch_pid_ pid exited..\n");
#endif
  ext_tsk();
}
void WSDdevfbExecute::_read_thread_(int ptr){
#ifdef PDEBUG
  printf("WSDdevfbExecute::_read_thread_.. start tid=%d\n",get_tid());
  dbprintf("WSDdevfbExecute::_read_thread_.. start tid=%d\n",get_tid());
#endif
  WSDdevfbExecute* _this = (WSDdevfbExecute*)ptr;
#ifdef PDEBUG
  printf("WSDdevfbExecute::_read_thread_ _this=0x%x\n",_this);
  dbprintf("WSDdevfbExecute::_read_thread_ _this=0x%x\n",_this);
#endif
  void* handle = _this->_handle;
//  pid_t pid = _this->_pid;

//printf("here1\n");
  while(1){
    if (_this->_handle != handle){
      break;
    }

    char buffer[1024];
#ifdef PDEBUG
  printf("WSDdevfbExecute::_read_thread_ _read_thread start read..\n");
  dbprintf("WSDdevfbExecute::_read_thread_ _read_thread start read..\n");
#endif
    int ret = fread(buffer,1,1023,(FILE*)_this->_handle);
#ifdef PDEBUG
  printf("WSDdevfbExecute::_read_thread_ _read_thread done read..\n");
  dbprintf("WSDdevfbExecute::_read_thread_ _read_thread done read..\n");
#endif
    if (ret > 0){
      buffer[ret] = 0;
      wai_sem(_this->_sem,T_FOREVER);
      _this->_string << buffer;
      sig_sem(_this->_sem);
    }
    if (ret <= 0){
      break;
    }
    _read_event_send_type data;
    data.obj = _this;
    data.id = _this->_id;

    MESSAGE* msg;
    msg = (MESSAGE*)buffer;
    msg->msg_type = MS_TYPE3;
    msg->msg_size = sizeof(data);
    memcpy(msg->msg_body.ANYMSG.msg_str,&data,sizeof(data));
    snd_msg(0,msg,WAIT);

#ifdef _NOTSH
#ifndef NO_THREAD
extern int _wake_up_event();
    _wake_up_event();
#endif
#endif

    if (ret < 1){
      break;
    }
  }
#ifdef PDEBUG
  printf("WSDdevfbExecute::_read_thread return\n");
  dbprintf("WSDdevfbExecute::_read_thread return\n");
#endif
  ext_tsk();
}

long WSDdevfbExecute::getId(){
  return _id;
}

#endif
