//
// Copyright (C) 1999-2009 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#include <WSCindexData.h>
#include <win/WSDwinFS.h>
#include <WSCstring.h>
#include <WSDenv.h>
#ifndef _WSWINCE
#include <unistd.h>
#include <sys/stat.h>
#endif
#include <WSClistData.h>
#include <WSClocaleSet.h>
#include <Winnetwk.h>

// w32shortcut class declaration start
#include <stdio.h>
#include <string.h>
#ifndef _WSWINCE
#include <errno.h>
#endif
class w32shortcut {
protected:
  unsigned long flags1;
  unsigned long attr;
  unsigned long flags2;

  char target[1024];

  char *name;
  FILE *fd;
  char *error_str;

#define HEADER_END 0x4c
  int getint( int off, int base, unsigned int len,
	      unsigned long *p, char *msg=NULL );
  int status;
  int relPath;
  int relPathLen;
  int unicode;
#define FLAGS1_SHELL_ITEM_ID_LIST 1
#define FLAGS1_FILE_OR_DIRECTORY  2
#define FLAGS1_DESC_STRING        4
#define FLAGS1_RELATIVE_PATH      8
#define FLAGS1_WORKING_DIRECTORY 16
#define FLAGS1_COMMAND_LINE_ARGS 32
#define FLAGS1_CUSTOM_ICON       64
#define FLAGS1_UNICODE          128

#define FLAGS2_LOCAL    1
#define FLAGS2_NETWORK  2

  // attribute bit

#define ATTR_READONLY  (1<<0)
#define ATTR_HIDDEN    (1<<1)
#define ATTR_SYSTEM    (1<<2)
#define ATTR_VOLUME    (1<<3)
#define ATTR_DIRECTORY (1<<4)
#define ATTR_ARCHIVE   (1<<5)
#define ATTR_DEVICE    (1<<6)
#define ATTR_NORMAL    (1<<7)
#define ATTR_TEMPORARY (1<<8)
#define ATTR_SPARSE    (1<<9)
#define ATTR_REPARSE   (1<<10)
#define ATTR_COMPRESSED (1<<11)
#define ATTR_OFFLINE   (1<<12)
#define ATTR_NO_INDEX  (1<<13)
#define ATTR_ENCRYPTED (1<<14)

public:
  w32shortcut(char *name);
  ~w32shortcut();
  int getStatus(){return status;}
  char *getError(){return error_str;}
  char *getTarget(){return target;}
  int targetIsRelativePath(){return relPath;}
  int relativePathLength(){return relPathLen;}
  int targetIsUnicode(){return unicode;}
  long getAttribute(){return attr;}
  long getFlags(){return flags1;}
  int isDir(){return (attr&ATTR_DIRECTORY)!=0;}
};
// w32shortcut class declaration end

WSDfileSystem* _win_fs_chandler(){
  return new WSDwinFS();
}
void WSGFinitWinFileSystem(){
static long _init = 0;
  if (_init != 0){
    return;
  }
  _init = 1;
extern void WSGFwinEnvInit();
  WSGFwinEnvInit();
  WSDfileSystem::setCreateInstanceHandler(_win_fs_chandler);
}
#ifndef NO_GLOBAL_CONSTRUCTORS
class _win_fs_init{
  public: _win_fs_init(){
    WSGFinitWinFileSystem();
  };
};
_win_fs_init _win_fs_init_execute;
#endif

WSDwinFS::WSDwinFS(){}
WSDwinFS::~WSDwinFS(){}

char* WSDwinFS::adjustFileName(char* src){
static char  buf[1024];

  WSCstring str( WSGIappEnvironment()->getPlaneString(src) );
  str.delHeadSpace();
  str.replaceString("\n","",0);
  str.replaceString("\\","/",0);

  strncpy(buf,str.getString(),1023);
  return buf;
}

long WSDwinFS::rename(char* old_name,char* new_name){
  WSCstring name1(old_name);
  WSCstring name2(new_name);
#ifndef _WSWINCE
#ifdef UNICODE
  remove_utf8(name2.getString(WS_EN_UTF8));
  long ret = rename_utf8(name1.getString(WS_EN_UTF8),
                      name2.getString(WS_EN_UTF8));
#else
  remove(name2.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()));
  long ret = ::rename(name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()),
                      name2.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()));
#endif
  if (ret == 0){
    return WS_NO_ERR;
  }
#else
  WSCushort* str1 = WSGFgetUCS2(name1.getString(),WS_EN_DEFAULT);
  WSCushort* str2 = WSGFgetUCS2(name2.getString(),WS_EN_DEFAULT);
  DeleteFile(str2);
  long ret = ::MoveFile(str1,str2);
  delete str1;
  delete str2;
  if (ret != 0){
    return WS_NO_ERR;
  }
#endif
  return WS_ERR;
}

long WSDwinFS::deleteFile(char* name){
  WSCstring fname(name);
#ifndef _WSWINCE
#ifdef UNICODE
  char* fn = fname.getString(WS_EN_UTF8);
  if (remove_utf8(fn) == 0){
    return WS_NO_ERR;
  }
#else
  char* fn = fname.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  if (remove(fn) == 0){
    return WS_NO_ERR;
  }
#endif
#else
  WSCushort* str1 = WSGFgetUCS2(name,WS_EN_DEFAULT);
  if (DeleteFile(str1)){
    delete str1;
    return WS_NO_ERR;
  }
  delete str1;
#endif
  return WS_ERR;
}

long WSDwinFS::createDir(char* name){
  WSCstring fname(name);
  char* fn = fname.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
#ifndef _WSWINCE
#ifdef UNICODE
  if (_mkdir_utf8(fname.getString(WS_EN_UTF8)) == 0){
    return WS_NO_ERR;
  }
#else
  if (_mkdir(fn) == 0){
    return WS_NO_ERR;
  }
#endif
#else
  WSCushort* str1 = WSGFgetUCS2(fn,WS_EN_DEFAULT);
  if (CreateDirectory(str1,NULL) != 0){
    delete str1;
    return WS_NO_ERR;
  }
  delete str1;
#endif
  return WS_ERR;
}

long WSDwinFS::deleteDir(char* name){
  WSCstring fname(name);
  char* fn = fname.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
#ifndef _WSWINCE
#ifdef UNICODE
  if (rmdir_utf8(fname.getString(WS_EN_UTF8)) == 0){
    return WS_NO_ERR;
  }
#else
  if (rmdir(fn) == 0){
    return WS_NO_ERR;
  }
#endif
#else
  WSCushort* str1 = WSGFgetUCS2(fn,WS_EN_DEFAULT);
  if (RemoveDirectory(str1) != 0){
    delete str1;
    return WS_NO_ERR;
  }
  delete str1;
#endif
  return WS_ERR;
}

#define WS_CANNOT 0
#define WS_DIR 1
#define WS_WRITABLE_FILE 2
#define WS_READONLY_FILE 3
long WSDwinFS::check(char* fname){
#ifndef _WSWINCE
  WSCstring name(fname);
  char* fn = name.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  struct _stat sret;
  long ret = 0;
  if (fn[0] != 0 && fn[1] == ':' && fn[2] == 0 ){
    ret = WS_FS_DRIVE;
    UINT bk = SetErrorMode(SEM_FAILCRITICALERRORS);
#ifndef UNICODE
    DWORD retv = GetFileAttributes(fn);
#else
    WSCushort* str1 = WSGFgetUCS2(name,WS_EN_DEFAULT);
    DWORD retv = GetFileAttributes((WCHAR*)str1);
    delete str1;
#endif
    SetErrorMode(bk);
    if (retv == 0xffffffff){
      ret |= WS_FS_DRIVE_NO_MEDIA;
    }
    return ret;
  }
#ifdef UNICODE
  if (_stat_utf8(name.getString(WS_EN_UTF8),&sret) == -1){
    return 0;
  }
#else
  if (_stat(fn,&sret) == -1){
    return 0;
  }
#endif
  if ((sret.st_mode & S_IFMT) == S_IFDIR){
    ret |= WS_FS_DIR;
  }
  if (sret.st_mode & S_IFREG){
    ret |= WS_FS_FILE;
  }
#ifdef UNICODE
  if ( _access_utf8(name.getString(WS_EN_UTF8),W_OK) == -1){
    ret |= WS_FS_READONLY;
  }
#else
  if ( _access(fn,W_OK) == -1){
    ret |= WS_FS_READONLY;
  }
#endif
  return ret;
#else
  return 0;
#endif
}
#if 0
int _name_cmp(char* name1, char* name2){
  WSCstring str1(name1);
  WSCstring str2(name2);
  str1.to_upper();
  str2.to_upper();
  return strcmp(str1.getString(),str2.getString());
}
#endif
WSClistData* WSDwinFS::showNetDir(char* dn,char* prvn){
//printf("WSDwinFS::showNetDir 0x%x 0x%x\n",dname,prvname);
  WSCstring name1(dn);
  char* dname = name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  WSCstring name2(prvn);
  char* prvname = name2.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  if (prvname != NULL && dname == NULL){
    return showDir(dn);
  }
  if (dname != NULL){
//printf("WSDwinFS::showNetDir dname=%s\n",dn);
  }
  if (prvname != NULL){
//printf("WSDwinFS::showNetDir prvname=%s\n",prvn);
  }
  NETRESOURCE  NetResource;
  NETRESOURCE  lpBuffer[128];
  HANDLE       lphEnum;
  DWORD        lpcCount;
  DWORD        lpBufferSize;
  DWORD        re, i;
#ifndef _WSWINCE
#ifndef UNICODE
  char  lpBuffer2[4096];
#else
  WSCushort* lpBuff=NULL;
  WSCushort* lpBuff2=NULL;
#endif
#else
  WSCushort* lpBuff=NULL;
  WSCushort* lpBuff2=NULL;
#endif

  lphEnum  = 0;
  lpcCount = 0xFFFFFFF;
  lpBufferSize = sizeof(lpBuffer);

  NetResource.dwScope		= RESOURCE_GLOBALNET;
  NetResource.dwType		= RESOURCETYPE_ANY;
  NetResource.dwDisplayType	= RESOURCEDISPLAYTYPE_DOMAIN;
  NetResource.dwUsage		= RESOURCEUSAGE_CONTAINER;
  if (dname != NULL ){
//    WSCstring str(dname);
//    str.replaceString("/","\\",0);
#ifndef _WSWINCE
#ifndef UNICODE
    strcpy(lpBuffer2,name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()));
    NetResource.lpRemoteName = lpBuffer2;
#else
    lpBuff = WSGFgetUCS2(name1.getString(),WS_EN_DEFAULT);
    NetResource.lpRemoteName = (WCHAR*)lpBuff;
#endif
#else
    lpBuff = WSGFgetUCS2(name1.getString(),WS_EN_DEFAULT);
    NetResource.lpRemoteName = lpBuff;
#endif
  }else{
#ifndef _WSWINCE
#ifndef UNICODE
    NetResource.lpRemoteName = dname;
#else
    lpBuff = WSGFgetUCS2(name1.getString(),WS_EN_DEFAULT);
    NetResource.lpRemoteName = (WCHAR*)lpBuff;
#endif
#else
    lpBuff = WSGFgetUCS2(name1.getString(),WS_EN_DEFAULT);
    NetResource.lpRemoteName = lpBuff;
#endif
  }
#ifndef _WSWINCE
#ifndef UNICODE
  NetResource.lpProvider		= prvname;
#else
  lpBuff2 = WSGFgetUCS2(name2.getString(),WS_EN_DEFAULT);
  NetResource.lpProvider		= (WCHAR*)lpBuff2;
#endif
#else
  lpBuff2 = WSGFgetUCS2(name2.getString(),WS_EN_DEFAULT);
  NetResource.lpProvider		= lpBuff2;
#endif
  re = WNetOpenEnum(RESOURCE_GLOBALNET, RESOURCETYPE_ANY, 0, &NetResource, &lphEnum);
  if (re != 0){
#ifndef _WSWINCE
#ifdef UNICODE
    if (lpBuff != NULL){
      delete lpBuff;
    }
    if (lpBuff2 != NULL){
      delete lpBuff2;
    }
#endif
#else
    if (lpBuff != NULL){
      delete lpBuff;
    }
    if (lpBuff2 != NULL){
      delete lpBuff2;
    }
#endif
    return NULL;
  }

  WSClistData* ret = NULL;
  for (;;) {
    re = WNetEnumResource(lphEnum, &lpcCount, &lpBuffer[0], &lpBufferSize);
    if (re == ERROR_NO_MORE_ITEMS){
      break;
    }else if (re != NO_ERROR){
      break;
    }
    for (i = 0;; i++) {
      if (i >= 128){
        break;
      }else if (i >= lpcCount){
        break;
      }

      WSCstring str;
#ifndef _WSWINCE
#ifndef UNICODE
      str.setString(lpBuffer[i].lpRemoteName,WSGIappLocaleSet()->getSystemLocaleEncoding());
#else
      char* tmpstr = WSGFgetString((WSCushort*)lpBuffer[i].lpRemoteName,WS_EN_DEFAULT);
      str.setString(tmpstr);
      delete tmpstr;
#endif
#else
      char* tmpstr = WSGFgetString(lpBuffer[i].lpRemoteName,WS_EN_DEFAULT);
      str.setString(tmpstr);
      delete tmpstr;
#endif
      str.replaceString("\\","/",0);
      if (dname != NULL ){
        if (!strncmp(dname,"//",2)){
          WSCstring tmp(dname);
          tmp << "/";
          str.replaceString(tmp,"",1);
        }else
        if (!strncmp(dname,"\\\\",2)){
          WSCstring tmp(dname);
          tmp.replaceString("\\\\","//",1);
          tmp << "/";
          str.replaceString(tmp,"",1);
        }
      }
      if (ret == NULL){
        ret = new WSClistData();
      }
      ret->add((void*)WSGFstrdup((char*)str));
    }
    break;
  }
  WNetCloseEnum(lphEnum);
#ifndef _WSWINCE
#ifdef UNICODE
    if (lpBuff != NULL){
      delete lpBuff;
    }
    if (lpBuff2 != NULL){
      delete lpBuff2;
    }
#endif
#else
    if (lpBuff != NULL){
      delete lpBuff;
    }
    if (lpBuff2 != NULL){
      delete lpBuff2;
    }
#endif
  return ret;
}
WSClistData* WSDwinFS::showDir(char* name){
  return showDirWithLink(name, NULL);
}
WSClistData* WSDwinFS::showDirWithLink(char* name,
				       WSCindexData** linklist){
dbprintf("WSDwinFS::showDirWithLink #%s#\n",name);
  WSClistData* ret = new WSClistData;
  if( linklist!=NULL ){
    if( *linklist!=NULL ){
      (*linklist)->clear();
    }else{
      *linklist=new WSCindexData;
    }
  }
  WIN32_FIND_DATA fd;
  WSCstring str;
  str.setString(name,WS_EN_DEFAULT);
#ifndef _WSWINCE
  if (!strcmp("",name)){
    DWORD drives = GetLogicalDrives();
    unsigned int i;
    for(i=0; i<sizeof(drives)*8;i++){
      char buf[3];
      if (drives & (1<<i)){
        buf[0] = 'A'+i;
        buf[1] = ':';
        buf[2] = 0;
        ret->add((void*)WSGFstrdup(buf));
      }
    }
    return ret;
  }
#endif
  //  WSCstring str0;
  //  str0=str;
  str << "/*";
  str.replaceString("/","\\",0);
  //  str0 << "/";
  //  str0.replaceString("/","\\",0);
dbprintf("WSDwinFS::showDir dname=#%s#\n",(char*)str);
#ifndef _WSWINCE
  long encoding = WSGIappLocaleSet()->getSystemLocaleEncoding();
  {
    WSCstring str2;
    str2.setString(name,WS_EN_DEFAULT);
    str2.replaceString("/","\\",0);
    UINT bk = SetErrorMode(SEM_FAILCRITICALERRORS);
#ifndef UNICODE
    DWORD retv = GetFileAttributes(str2.getString(encoding));
#else
    WSCushort* tmpstr = WSGFgetUCS2(str2,WS_EN_DEFAULT);
    DWORD retv = GetFileAttributes((WCHAR*)tmpstr);
    delete tmpstr;
#endif
    SetErrorMode(bk);
    if (retv == 0xffffffff){
        return ret;
    }
  }
#ifndef UNICODE
  HANDLE hFind = FindFirstFile(str.getString(encoding), &fd);
#else
  WSCushort* tmpstr = WSGFgetUCS2(str,WS_EN_DEFAULT);
  HANDLE hFind = FindFirstFile((WCHAR*)tmpstr,&fd);
  delete tmpstr;
#endif
#else
  WSCushort* tmpstr = WSGFgetUCS2(str,WS_EN_DEFAULT);
  HANDLE hFind = FindFirstFile(tmpstr,&fd);
  delete tmpstr;
#endif
  while( hFind != INVALID_HANDLE_VALUE ){
    WSCbool isDir=False;
    WSCstring *linkTarget;

    linkTarget=NULL;
    if ( (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0){
      isDir=True;
    }else{
#ifndef _WSWINCE
      if( linklist && *linklist ){
	int len;
#ifndef UNICODE
	len=strlen(fd.cFileName);
	if( len>=4 && !stricmp( fd.cFileName+len-4, ".LNK" ) ){
	  WSCstring full;
	  //	  full=str0;
	  full=str;full.cutString(full.getChars()-1);
	  full.addString(fd.cFileName,encoding);
	  w32shortcut *s=new w32shortcut(full.getString(encoding));
	  if( s!=NULL && s->getStatus()!=0 ){
	    if( s->isDir() ){
	      isDir=True;
	      linkTarget=new WSCstring(s->getTarget(),encoding);
	      linkTarget->replaceString("\\","/",0);
	    }
	  }
	  delete s;
	}
#else
	char* tstr = WSGFgetString((WSCushort*)fd.cFileName,WS_EN_DEFAULT);
	len=strlen(tstr);
	if( len>=4 && !stricmp( tstr+len-4, ".LNK" ) ){
	  WSCstring full;
	  //	  full=str0;
	  full=tstr;
          full.cutString(full.getChars()-1);
	  full.addString(fd.cFileName,encoding);
	  w32shortcut *s=new w32shortcut(full.getString(WS_EN_UTF8));
	  if( s!=NULL && s->getStatus()!=0 ){
	    if( s->isDir() ){
	      isDir=True;
	      linkTarget=new WSCstring(s->getTarget(),encoding);
	      linkTarget->replaceString("\\","/",0);
	    }
	  }
	  delete s;
	}
	delete tstr;
#endif
      }
#endif
    }
    if( isDir ){
#ifndef _WSWINCE
#ifndef UNICODE
       if (!strcmp(fd.cFileName,".") || !strcmp(fd.cFileName,"..")){
         // reject...
       }else{
	 WSCstring str;//
         str.setString(fd.cFileName,encoding);
#else
       char* tmpst = WSGFgetString((WSCushort*)fd.cFileName,WS_EN_DEFAULT);
       WSCstring tmps(tmpst);
       delete tmpst;
       if (!strcmp(tmps.getString(),".") || !strcmp(tmps.getString(),"..")){
         // reject...
       }else{
	 WSCstring str(tmps);//
#endif
         ret->add((void*)WSGFstrdup(str.getString(WS_EN_DEFAULT)));
	 if( linkTarget ){
	   (*linklist)->setData(str.getString(WS_EN_DEFAULT),
		(void*)linkTarget->getString((WS_EN_DEFAULT)));
	 }
       }
#else
       char* tmpst = WSGFgetString(fd.cFileName,WS_EN_DEFAULT);
       WSCstring tmps(tmpst);
       delete tmpst;
       if (!strcmp(tmps.getString(),".") || !strcmp(tmps.getString(),"..")){
         // reject...
       }else{
         ret->add((void*)WSGFstrdup(tmps.getString(WS_EN_DEFAULT)));
       }
#endif
    }
    WSCbool exist = FindNextFile(hFind,&fd);
    if (exist == False){
      FindClose(hFind);
      break;
    }
  }
  WSGFsortCharList(ret);
  return ret;
}
WSClistData* WSDwinFS::showFile(char* name){
  return showFileWithLink( name, NULL );
}

WSClistData* WSDwinFS::showFileWithLink(char* name,
				WSCindexData **linklist){
  WSClistData* ret = new WSClistData();
dbprintf("WSDwinFS::showFileWithLink #%s#\n",name);
  if (!strcmp(name,"")){
    return ret;
  }
  if( linklist!=NULL ){
    if( *linklist!=NULL ){
      (*linklist)->clear();
    }else{
      *linklist=new WSCindexData;
    }
  }
  WIN32_FIND_DATA fd;
  WSCstring str;
  //  WSCstring str0;
  str.setString(name,WS_EN_DEFAULT);
  //  str0=str;
  str << "/*";
  str.replaceString("/","\\",0);
  //  str0 << "/";
  //  str0.replaceString("/","\\",0);
#ifndef _WSWINCE
  long encoding = WSGIappLocaleSet()->getSystemLocaleEncoding();
  {
    WSCstring str2;
    str2.setString(name,WS_EN_DEFAULT);
    str2.replaceString("/","\\",0);
    UINT bk = SetErrorMode(SEM_FAILCRITICALERRORS);
#ifndef UNICODE
    DWORD retv = GetFileAttributes(str2.getString(encoding));
#else
    WSCushort* tmpstr = WSGFgetUCS2(str2,WS_EN_DEFAULT);
    DWORD retv = GetFileAttributes((WCHAR*)tmpstr);
    delete tmpstr;
#endif
    SetErrorMode(bk);
    if (retv == 0xffffffff){
        return ret;
    }
  }
#ifndef UNICODE
  HANDLE hFind = FindFirstFile(str.getString(encoding), &fd);
#else
  WSCushort* tmpstr = WSGFgetUCS2(str,WS_EN_DEFAULT);
  HANDLE hFind = FindFirstFile((WCHAR*)tmpstr,&fd);
  delete tmpstr;
#endif
#else
  WSCushort* tmpstr = WSGFgetUCS2(str,WS_EN_DEFAULT);
  HANDLE hFind = FindFirstFile(tmpstr,&fd);
dbprintf("WSDwinFS::showFile #%s# hFind=0x%x\n",name,hFind);
  delete tmpstr;
#endif
  while( hFind != INVALID_HANDLE_VALUE ){
    WSCbool isFile=True;
    WSCstring *linkTarget;

    linkTarget=NULL;
    if ( (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0){
      isFile=False;
    }else{
#ifndef _WSWINCE
      if( linklist && *linklist ){
	int len;
#ifndef UNICODE
	len=strlen(fd.cFileName);
	if( len>=4 && !stricmp( fd.cFileName+len-4, ".LNK" ) ){
	  WSCstring full;
	  //full=str0;
	  full=str;full.cutString(full.getChars()-1);
	  full.addString(fd.cFileName,encoding);
	  w32shortcut *s=new w32shortcut(full.getString(encoding));
	  if( s!=NULL && s->getStatus()!=0 ){
	    if( s->isDir() ){
	      isFile=False;
	    }else{
	      linkTarget=new WSCstring(s->getTarget(),encoding);
	      linkTarget->replaceString("\\","/",0);
	    }
	    delete s;
	  }
	}
#else
	char* tstr = WSGFgetString((WSCushort*)fd.cFileName,WS_EN_DEFAULT);
	len=strlen(tstr);
	if( len>=4 && !stricmp( tstr+len-4, ".LNK" ) ){
	  WSCstring full;
	  //	  full=str0;
	  full=tstr;full.cutString(full.getChars()-1);
	  full.addString(fd.cFileName,encoding);
	  w32shortcut *s=new w32shortcut(full.getString(WS_EN_UTF8));
	  if( s!=NULL && s->getStatus()!=0 ){
	    if( s->isDir() ){
	      isFile=False;
	    }else{
	      linkTarget=new WSCstring(s->getTarget(),encoding);
	      linkTarget->replaceString("\\","/",0);
	    }
	    delete s;
	  }
	}
#endif
      }
#endif
    }
    if( isFile ){
#ifndef _WSWINCE
      WSCstring str;
#ifndef UNICODE
      str.setString(fd.cFileName,WSGIappLocaleSet()->getSystemLocaleEncoding());
#else
      char* tstr = WSGFgetString((WSCushort*)fd.cFileName,WSGIappLocaleSet()->getSystemLocaleEncoding());
      str.setString(tstr,WSGIappLocaleSet()->getSystemLocaleEncoding());
      delete tstr;
#endif
      ret->add((void*)WSGFstrdup(str.getString(WS_EN_DEFAULT)));
      if( linkTarget ){
	(*linklist)->setData(str.getString(WS_EN_DEFAULT),
	     (void*)linkTarget->getString((WS_EN_DEFAULT)));
      }
#else
      char* tmpst = WSGFgetString(fd.cFileName,WS_EN_DEFAULT);
      ret->add((void*)tmpst);
//      ret->add((void*)WSGFstrdup(tmpst));
//      delete tmpst;
#endif
    }
    WSCbool exist = FindNextFile(hFind,&fd);
    if (exist == False){
      FindClose(hFind);
      break;
    }
  }
  WSGFsortCharList(ret);
  return ret;
}

WSCstring *WSDwinFS::readLink(char*linkname){
  WSCstring *p=new WSCstring(linkname);
#ifdef UNICODE
  w32shortcut s(p->getString(WS_EN_UTF8));
#else
  w32shortcut s(p->getString(WSGIappLocaleSet()
			     ->getSystemLocaleEncoding()));
#endif
  if( s.getStatus()!=0 ){
    char *target=s.getTarget();
    if( target ){
      p->setString(target,
		    WSGIappLocaleSet()->getSystemLocaleEncoding());
      p->replaceString("\\","/",0);
      return p;
    }
  }
  return NULL;
}

// w32shortcut definition start
static const unsigned char magic[]={
  'L',0,0,0,0x01,
  0x14,0x02,0x00,
  0,0,0,0,
  0xC0,0,0,0,
  0,0,0,0x46};

int
w32shortcut::getint( int off, int base, unsigned int len,
		     unsigned long *p, char *msg ){
    unsigned char buff[1024];
    unsigned long n=0;
    int r;
    int i;

    if( len>sizeof(buff) )return -1;

    r=fseek( fd, off, base );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      return -1;
    }

    r=fread( buff, 1, len, fd );
    if( r<0 || (unsigned int)r<len ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      return -1;
    }
    for( i=len-1 ; i>=0 ; --i ){
	n=n*256+buff[i];
    }
    *p=n;
    return r;
}

#ifdef MSW
int
bcmp( const unsigned char *a, const unsigned char *b, int len ){
  int i;
  for( i=0 ; i<len && a[i]==b[i] ; i++ );
  return i>=len?0:a[i]>b[i]?1:-1;
}
#endif

w32shortcut::~w32shortcut(){
  delete name;
}

w32shortcut::w32shortcut( char *linkname ){
  unsigned char buff[1024];
  unsigned int i;

  status=0;
  error_str="";
  unicode=0;
  relPath=0;
  relPathLen=0;

  name=new char[strlen(linkname)+1];

  if( name==NULL ){
    error_str="insufficient memory";
    return;
  }

  strcpy(name,linkname);
#ifdef MSW
#ifdef UNICODE
  fd=fopen_utf8(name,"rb");
#else
  fd=fopen(name,"rb");
#endif
#else
#ifdef UNICODE
  fd=fopen_utf8(name,"r");
#else
  fd=fopen(name,"r");
#endif
#endif
  if( fd==NULL ){
#ifndef _WSWINCE
    error_str=strerror(errno);
#endif
    //fprintf(stderr,"%s (%s)\n",name,error_str);
    return;
  }

  int l;
  int r;

  l=sizeof( magic );
  r=fread( buff, 1, l, fd );

  if( r<l || bcmp( buff, magic, l )!=0 ){
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose( fd );
      //fprintf(stderr,"no magic(r=%d:%s)\n",r,error_str);
      return;
    }
    fclose( fd );
    error_str="no magic";
    return;
  }

  r=getint( 0, SEEK_CUR, 4, &flags1 );

  if( r<0 ){
#ifndef _WSWINCE
    error_str=strerror(errno);
#endif
    fclose(fd);
    //fprintf(stderr,"no flag1(r=%d:%s)\n",r,error_str);
    return;
  }
  //fprintf(stderr,"%s:flags1=%08x:\n", name, flags1);

  r=getint( 0, SEEK_CUR, 4, &attr );

  if( r<0 ){
#ifndef _WSWINCE
    error_str=strerror(errno);
#endif
    fclose(fd);
    //fprintf(stderr,"no attr(r=%d:%s)\n",r,error_str);
    return;
  }
  //fprintf(stderr,"%s:attr=%08x:\n", name, attr);

  r=fseek( fd, HEADER_END, SEEK_SET );
  if( r<0 ){
#ifndef _WSWINCE
    error_str=strerror(errno);
#endif
    fclose(fd);
    //fprintf(stderr,"header skip error(r=%d:%s)\n",r,error_str);
    return;
  }

  if( flags1 & FLAGS1_SHELL_ITEM_ID_LIST ){
    unsigned long len;
    r=getint( 0, SEEK_CUR, 2, &len );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"shell item length error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
    r=fseek( fd, len, SEEK_CUR );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"shell item skip error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
  }

  if( flags1 & FLAGS1_FILE_OR_DIRECTORY ){
    long off0;
    unsigned long offhead, offheadoff;
    unsigned long offtail;
    unsigned long len2;

    off0=ftell( fd );

    if( off0<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"file_or_directory error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }

    r=getint( 0, SEEK_CUR, 4, &len2 );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"file_or_directory length error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }

    r=getint( off0+8, SEEK_SET, 4, &flags2 );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"file_or_directory flag2 error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }

    if( flags2&(FLAGS2_LOCAL|FLAGS2_NETWORK)==0 ){
      fclose(fd);
      error_str="illegal location flag";
      /*	fprintf(stderr,"%s:illegal location flag %08x\n",
		name,flags2);*/
      return;
    }
    if( flags2&FLAGS2_LOCAL ){
      offheadoff=0x10;
    }else{
      offheadoff=0x14;
    }
    r=getint( off0+offheadoff, SEEK_SET, 4, &offhead );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,
	"file_or_directory head location error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
    r=getint( off0+0x18, SEEK_SET, 4, &offtail );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,
	"file_or_directory tail location error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
    if( !(flags2&FLAGS2_LOCAL) ){
      /* to skip header the part of network volume table */
      offhead+=0x14;
    }
    //(global-set-key "\C-Cf" 'word-search-forward)

    i=0;
    target[i]=0;
    r=fseek( fd, off0+offhead, SEEK_SET );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,
	"file_or_directory head part seek error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
    while( (r=getc(fd))!=EOF ){
      if( i<sizeof(target) ){
	target[i++]=r;
	if( r==0 )break;
      }else{
	fclose(fd);
	error_str="link target too long\n";
	//fprintf(stderr,"link target too long\n");
	return;
      }
    }

    if( !(flags2&FLAGS2_LOCAL) ){
      if( i>0 ){
	if( i<sizeof(target) ){
	  target[i-1]='\\';
	  target[i++]=0;
	}else{
	  fclose(fd);
	  error_str="link target too long\n";
	  return;
	}
      }
    }

    r=fseek( fd, off0+offtail, SEEK_SET );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,
	"file_or_directory tail part seek error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }

    if( i>0 )i--;
    while( (r=getc(fd))!=EOF ){
      if( i<sizeof(target) ){
	target[i++]=r;
	if( r==0 )break;
      }else{
	fclose(fd);
	error_str="link target too long";
	//fprintf(stderr,"link target too long\n");
	return;
      }
    }
    //fprintf(fdo,"target='%s'\n",target);

    r=fseek( fd, off0+len2, SEEK_SET );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,
	"file_or_directory end seek error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
    fclose(fd);
    status=1;
    return;
  }// FILE_OR_DIRECTORY
#if 1
  if( flags1&FLAGS1_DESC_STRING ){
    unsigned long off0, len;

    off0=ftell( fd );

    if( off0<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"description error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }

    r=getint( 0, SEEK_CUR, 4, &len );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"description length error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }

    if( flags1 & FLAGS1_UNICODE ){
      r=fseek( fd, len*2, SEEK_CUR );/* *2 for unicode */
    }else{
      r=fseek( fd, len, SEEK_CUR );/* for not unicode */
    }
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"description skip error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
  }
  if( flags1&FLAGS1_RELATIVE_PATH ){
    unsigned long off0, len;

    off0=ftell( fd );

    if( off0<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"relative path error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }

    r=getint( 0, SEEK_CUR, 4, &len );
    if( r<0 ){
#ifndef _WSWINCE
      error_str=strerror(errno);
#endif
      fclose(fd);
      /*fprintf(stderr,"relative path length error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
    if( flags1 & FLAGS1_UNICODE ){
      len*=2;
    }
    if( len>sizeof(target) ){
      fclose(fd);
      error_str="link target relative path too long";
      return;
    }
    r=fread( target, 1, len, fd );
    if( r<0 || (unsigned int)r<len ){
      if( r<0 ){
#ifndef _WSWINCE
	error_str=strerror(errno);
#endif
      }else{
	error_str="relative path read error";
      }
      fclose(fd);
      /*fprintf(stderr,"relative path skip error(r=%d:%s)\n",
	r,error_str);*/
      return;
    }
    relPath=1;
    relPathLen=len;
    if( flags1 & FLAGS1_UNICODE ){
      unicode=1;
    }
    fclose(fd);
    status=1;
    return;
  }
  /* WORKING_DIRECTORY */
  /* COMMAND_LINE_ARGUMENTS */
  /* CUSTOM_ICON */
  fclose(fd);
  status=0;
#endif
}
// w32shortcut definition end
