//
// Copyright (C) 1999-2008 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WScom.h>
#ifndef NO_COMM
#include <x11/WSDunixComm.h>
#include <stdio.h>
#include <fcntl.h>
#include <WSClocaleSet.h>
#include <unistd.h>
#include <termios.h>

WSMFclassInit(WSDunixComm,WSDcomm);

WSDcomm* _unix_comm_chandler(){
  return new WSDunixComm;
}

void WSGFinitializeUnixComm(){
    WSDcomm::setCreateInstanceHandler((void*)_unix_comm_chandler);
}
#ifndef NO_GLOBAL_CONSTRUCTORS
class _unix_comm_init {
  public: _unix_comm_init(){
    WSGFinitializeUnixComm();
  };
};
_unix_comm_init  _unix_comm_init_execute;
#endif

WSDunixComm::WSDunixComm(){
  _fp = -1;
}

WSDunixComm::~WSDunixComm(){
  close();
}

long WSDunixComm::open(char* name,int mode){
  _com_name.setString(name);
  _com_name.delHeadSpace();
  int md =0;
  if (mode == WS_READ){
    md = O_RDONLY | O_NOCTTY;
  }else
  if (mode == WS_WRITE){
    md = O_WRONLY | O_NOCTTY;
  }else{
    md = O_RDWR | O_NOCTTY;
  }
  char* fn = _com_name.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  _fp = ::open(fn,md,0777);
  if (_fp > -1){
    if (_baud != WS_BAUD_DEFAULT){
      int baud = B9600;
      switch(_baud){
        case WS_B115200:{
          baud = B115200;
          break;
        }
        case WS_B57600:{
          baud = B57600;
          break;
        }
        case WS_B38400:{
          baud = B38400;
          break;
        }
        case WS_B19200:{
          baud = B19200;
          break;
        }
        default:{
          baud = B9600;
          break;
        }
      } 
      struct  termios rs_arg;
      if (tcgetattr(_fp, &rs_arg) == -1){
        return WS_ERR;
      }

      cfsetospeed(&rs_arg, baud);
      cfsetispeed(&rs_arg, 0);
      rs_arg.c_cflag |= CRTSCTS | CS8 | CREAD;
      rs_arg.c_lflag = 0;
      rs_arg.c_cc[VTIME]    = 0;
      rs_arg.c_cc[VMIN]     = 1;
      rs_arg.c_iflag = IGNPAR;
      rs_arg.c_oflag = 0;
      tcflush(_fp, TCIFLUSH);
      if (tcsetattr(_fp, TCSANOW, &rs_arg) == -1){
        return WS_ERR;
      }
    }
    return WS_NO_ERR;
  }else{
    _fp = -1;
    return WS_ERR;
  }
}

long WSDunixComm::close(){
  if (_fp > -1){
    ::close(_fp);
    _fp = -1;
  }
  return WS_NO_ERR;
}
long WSDunixComm::read(void* buf,long len){
  if (_fp == -1){
    return 0;
  }
  return ::read(_fp,buf,len);
}
long WSDunixComm::write(void* buf,long len){
  if (_fp == -1){
    return 0;
  }
  return ::write(_fp,buf,len);
}
#endif //NO_COMM

