//
// Copyright (C) 1999-2008 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <WSCstring.h>
#include <WSDenv.h>
#include <WSClistData.h>
#include <WSCindexData.h>
#include <x11/WSDunixFS.h>
#include <unistd.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <WSClocaleSet.h>

// w32shortcut class declaration start
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

class w32shortcut {
protected:
  unsigned long flags1;
  unsigned long attr;
  unsigned long flags2;

  char target[1024];

  char *name;
  FILE *fd;
  char *error_str;

  static const char HEADER_END=0x4C;
  int getint( int off, int base, unsigned int len,
	      unsigned long *p, char *msg=NULL );
  int status;
  int relPath;
  int relPathLen;
  int unicode;
  static const unsigned long FLAGS1_SHELL_ITEM_ID_LIST=1;
  static const unsigned long FLAGS1_FILE_OR_DIRECTORY =2;
  static const unsigned long FLAGS1_DESC_STRING       =4;
  static const unsigned long FLAGS1_RELATIVE_PATH     =8;
  static const unsigned long FLAGS1_WORKING_DIRECTORY=16;
  static const unsigned long FLAGS1_COMMAND_LINE_ARGS=32;
  static const unsigned long FLAGS1_CUSTOM_ICON      =64;
  static const unsigned long FLAGS1_UNICODE         =128;

  static const unsigned long FLAGS2_LOCAL    =1;
  static const unsigned long FLAGS2_NETWORK  =2;

  // attribute bit

  static const unsigned long ATTR_READONLY  =1<<0;
  static const unsigned long ATTR_HIDDEN    =1<<1;
  static const unsigned long ATTR_SYSTEM    =1<<2;
  static const unsigned long ATTR_VOLUME    =1<<3;
  static const unsigned long ATTR_DIRECTORY =1<<4;
  static const unsigned long ATTR_ARCHIVE   =1<<5;
  static const unsigned long ATTR_DEVICE    =1<<6;
  static const unsigned long ATTR_NORMAL    =1<<7;
  static const unsigned long ATTR_TEMPORARY =1<<8;
  static const unsigned long ATTR_SPARSE    =1<<9;
  static const unsigned long ATTR_REPARSE   =1<<10;
  static const unsigned long ATTR_COMPRESSED=1<<11;
  static const unsigned long ATTR_OFFLINE   =1<<12;
  static const unsigned long ATTR_NO_INDEX  =1<<13;
  static const unsigned long ATTR_ENCRYPTED =1<<14;

public:
  w32shortcut(char *name);
  ~w32shortcut();
  int getStatus(){return status;}
  char *getError(){return error_str;}
  char *getTarget(){return target;}
  int targetIsRelativePath(){return relPath;}
  int relativePathLength(){return relPathLen;}
  int targetIsUnicode(){return unicode;}
  long getAttribute(){return attr;}
  long getFlags(){return flags1;}
  int isDir(){return (attr&ATTR_DIRECTORY)!=0;}
};
// w32shortcut class declaration end

WSDfileSystem* _unix_fs_chandler(){
  return new WSDunixFS();
}
void WSGFinitUnixFileSystem(){
extern void WSGFunixEnvInit();
    WSGFunixEnvInit();
    WSDfileSystem::setCreateInstanceHandler(_unix_fs_chandler);
}
#ifndef NO_GLOBAL_CONSTRUCTORS
class _unix_fs_init{
  public: _unix_fs_init(){
    WSGFinitUnixFileSystem();
  };
};
_unix_fs_init _unix_fs_init_execute;
#endif

WSDunixFS::WSDunixFS(){}
WSDunixFS::~WSDunixFS(){}

char* WSDunixFS::adjustFileName(char* src){
static char  buf[1024];
//printf("WSDunixFS::adjustFileName(%s)\n",src);
  WSCstring str( WSGIappEnvironment()->getPlaneString(src));
  str.delHeadSpace();
  str.replaceString("\n","",0);
  str.replaceString("\\","/",0);

  char* fn = str.getString();
  if (fn[0] != 0 && fn[1] == ':'){
    str.deleteChars(0,2);
  }

  strncpy(buf,str.getString(),1023);
//printf("WSDunixFS::adjustFileName ->%s\n",buf);
  return buf;
}

static long unix_rename(char* old_name,char* new_name){
  if (rename(old_name,new_name) == 0){
    return WS_NO_ERR;
  }
  return WS_ERR;
}

long WSDunixFS::rename(char* old_name,char* new_name){
  WSCstring name1(old_name);
  WSCstring name2(new_name);
#ifdef MacOS
  if (WSGIappLocaleSet()->getSystemLocaleEncoding() == WS_EN_UTF8){
extern char* utf8_to_macutf8(char*);
    char* mstr1 = utf8_to_macutf8(
          name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()));
    char* mstr2 = utf8_to_macutf8(
          name2.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()));
    long ret;
    if (mstr1 != NULL && mstr2 != NULL){
      ret = unix_rename(mstr1,mstr2);
    }
    if (mstr1 != NULL){
      delete mstr1;
    }
    if (mstr2 != NULL){
      delete mstr2;
    }
    return ret;
  }else{
    return unix_rename(
     name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()),
     name2.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()));
  }
#else
  return unix_rename(
     name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()),
     name2.getString(WSGIappLocaleSet()->getSystemLocaleEncoding()));
#endif
}

long WSDunixFS::deleteFile(char* name){
  WSCstring name1(name);
  char* fn = name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
#ifdef MacOS
  if (WSGIappLocaleSet()->getSystemLocaleEncoding() == WS_EN_UTF8){
extern char* utf8_to_macutf8(char*);
    char* fn2 = utf8_to_macutf8(fn);
    if (fn2 != NULL){
      if (remove(fn2) == 0){
        delete fn2;
        return WS_NO_ERR;
      }
      delete fn2;
    }else{
      if (remove(fn) == 0){
        return WS_NO_ERR;
      }
    }
  }else{
    if (remove(fn) == 0){
      return WS_NO_ERR;
    }
  }
#else
  if (remove(fn) == 0){
    return WS_NO_ERR;
  }
#endif
  return WS_ERR;
}

long WSDunixFS::createDir(char* name){
  WSCstring name1(name);
  char* fn = name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
#ifdef MacOS
  if (WSGIappLocaleSet()->getSystemLocaleEncoding() == WS_EN_UTF8){
extern char* utf8_to_macutf8(char*);
    char* fn2 = utf8_to_macutf8(fn);
    if (fn2 != NULL){
      if (mkdir(fn2,0777) == 0){
        delete fn2;
        return WS_NO_ERR;
      }
      delete fn2;
    }else{
      if (mkdir(fn,0777) == 0){
        return WS_NO_ERR;
      }
    }
  }else{
    if (mkdir(fn,0777) == 0){
      return WS_NO_ERR;
    }
  }
#else
  if (mkdir(fn,0777) == 0){
    return WS_NO_ERR;
  }
#endif
  return WS_ERR;
}

long WSDunixFS::deleteDir(char* name){
  WSCstring name1(name);
  char* fn = name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
#ifdef MacOS
  if (WSGIappLocaleSet()->getSystemLocaleEncoding() == WS_EN_UTF8){
extern char* utf8_to_macutf8(char*);
    char* fn2 = utf8_to_macutf8(fn);
    if (fn2 != NULL){
      if (rmdir(fn2) == 0){
        delete fn2;
        return WS_NO_ERR;
      }
      delete fn2;
    }else{
      if (rmdir(fn) == 0){
        return WS_NO_ERR;
      }
    }
  }else{
    if (rmdir(fn) == 0){
      return WS_NO_ERR;
    }
  }
#else
  if (rmdir(fn) == 0){
    return WS_NO_ERR;
  }
#endif
  return WS_ERR;
}
#if 0
#define WS_CANNOT 0
#define WS_DIR 1
#define WS_WRITABLE_FILE 2
#define WS_READONLY_FILE 3
static long chk_file(char* file){
  struct stat sret;
  if (stat(file,&sret) == -1){
    return 0;
  }else if ((sret.st_mode & S_IFMT) == S_IFDIR){
    return WS_DIR;
  }else if (!(sret.st_mode & S_IFREG) || access(file,W_OK) == -1){
    return WS_READONLY_FILE;
  }
  return WS_WRITABLE_FILE;
}
#endif
long WSDunixFS::check(char* fname){
  WSCstring name1(fname);
  char* fn = name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  struct stat sret;
  long ret = 0;
#ifdef MacOS
  char* fn2 = NULL;
  if (WSGIappLocaleSet()->getSystemLocaleEncoding() == WS_EN_UTF8){
extern char* utf8_to_macutf8(char*);
    fn2 = utf8_to_macutf8(fn);
  }else{
    fn2 = WSGFstrdup(fn);
  }
  if (fn2 == NULL){
    return 0;
  }
  if (stat(fn2,&sret) == -1){
    delete fn2;
    return 0;
  }
  if ((sret.st_mode & S_IFMT) == S_IFDIR){
    ret |= WS_FS_DIR;
  }
  if (sret.st_mode & S_IFREG){
    ret |= WS_FS_FILE;
  }
  if ( access(fn,W_OK) == -1){
    ret |= WS_FS_READONLY;
  }
  delete fn2;
#else
  if (stat(fn,&sret) == -1){
    return 0;
  }
  if ((sret.st_mode & S_IFMT) == S_IFDIR){
    ret |= WS_FS_DIR;
  }
  if (sret.st_mode & S_IFREG){
    ret |= WS_FS_FILE;
  }
  if ( access(fn,W_OK) == -1){
    ret |= WS_FS_READONLY;
  }
#endif 
  return ret;
}





WSClistData* WSDunixFS::showDir(char* dname){
  return showDirWithLink(dname,NULL);
}

WSClistData* WSDunixFS::showDirWithLink(char* dname,
					WSCindexData** linklist){
  long encoding = WSGIappLocaleSet()->getSystemLocaleEncoding();

  if( linklist!=NULL ){
    if( *linklist!=NULL ){
      (*linklist)->clear();
    }else{
      *linklist=new WSCindexData;
    }
  }

  struct stat sret;
  dirent* finddata;
  WSCstring name1(dname);
  char* dn = name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  DIR* dptr = opendir(dn);
//printf("dname=#%s#\n",dn);
  if (dptr == NULL){
//printf("dname=%s return NULL\n",dname);
    return NULL;
  }
  WSClistData* list = new WSClistData;
  seekdir(dptr,0);
  while( (finddata = readdir(dptr)) != NULL){
    WSCbool isDir=False;
    WSCstring *linkTarget;

    if (!strcmp(finddata->d_name,"..")){
      continue;
    }
    if (!strcmp(finddata->d_name,".")){
      continue;
    }
    WSCstring tmp2(finddata->d_name,encoding);
    WSCstring fname;
    fname << name1 << "/" << tmp2;
    int ret = stat(fname.getString(encoding),&sret);
//printf(" +-#%s# #%s# ret=%d mode=0x%x\n",finddata->d_name,fname.getString(encoding),ret,sret.st_mode);
    if (ret < 0){
      continue;
    }
//printf("sret.st_mode=0x%x dir=0x%x dn=%s\n",sret.st_mode,S_IFDIR,finddata->d_name);
    linkTarget=NULL;
    if ((sret.st_mode & S_IFMT) == S_IFDIR){
	isDir=True;
    }else if( linklist && *linklist ){
      int len;
      len=strlen(finddata->d_name);
      if( len>=4 && !strcasecmp( finddata->d_name+len-4, ".LNK" ) ){
	w32shortcut *s=new w32shortcut( fname.getString(encoding) );
	if( s!=NULL && s->getStatus()!=0 ){
	  if( s->isDir() ){
	    isDir=True;
	    linkTarget=new WSCstring(s->getTarget(),_link_encoding);
	    linkTarget->replaceString("\\","/",0);
	  }
	  delete s;
	}
      }
    }
    if ( isDir ){
      WSCstring tmp;
#ifdef MacOS
      if (WSGIappLocaleSet()->getSystemLocaleEncoding() == WS_EN_UTF8){
extern char* macutf8_to_utf8(char*);
        char* t2 = macutf8_to_utf8(finddata->d_name);
        if (t2 != NULL){
          tmp.setString(t2,encoding);
          delete t2;
        }else{
          tmp.setString(finddata->d_name,encoding);
        }
      }else{
        tmp.setString(finddata->d_name,encoding);
      }
#else
      tmp.setString(finddata->d_name,encoding);
#endif
      list->add((void*)WSGFstrdup(tmp.getString()));
      if( linkTarget ){
	  (*linklist)->setData(tmp.getString(),
			       (void*)linkTarget->getString((WS_EN_DEFAULT)));
      }
    }
  }
//printf(" done\n");
  closedir(dptr);
  WSGFsortCharList(list);
  return list;
#if 0
  char buf[8192];
  sprintf(buf,"/bin/ls %s",dname);

  FILE* pp;
  if (!(pp = popen(buf,"r"))){
    return NULL;
  }
  WSClistData* list = new WSClistData;

  while(fgets(buf,sizeof(buf),pp)){
    WSCstring str(buf);
    str.delString("\n",0);
    WSCstring str2(dname);
    str2.addString("/");
    str2.addString(str);
    if (chk_file(str2.getString()) == WS_DIR){
      list->add((void*)WSGFstrdup(str.getString()));
    }
  }
  pclose(pp);
  return list;
#endif
}
WSClistData* WSDunixFS::showFile(char* dname){
  return showFileWithLink( dname, NULL );
}
WSClistData* WSDunixFS::showFileWithLink(char* dname,
				 WSCindexData **linklist){
  long encoding = WSGIappLocaleSet()->getSystemLocaleEncoding();

  if( linklist!=NULL ){
    if( *linklist!=NULL ){
      (*linklist)->clear();
    }else{
      *linklist=new WSCindexData;
    }
  }

  struct stat sret;
  dirent* finddata;
  WSCstring name1(dname);
  char* dn = name1.getString(WSGIappLocaleSet()->getSystemLocaleEncoding());
  DIR* dptr = opendir(dn);
  if (dptr == NULL){
    return NULL;
  }
  WSClistData* list = new WSClistData;
  while( (finddata = readdir(dptr)) != NULL){
    WSCbool isFile=True;
    WSCstring *linkTarget;

    if (!strcmp(finddata->d_name,"..")){
      continue;
    }
    if (!strcmp(finddata->d_name,".")){
      continue;
    }
    WSCstring tmp2(finddata->d_name,encoding);
    WSCstring fname;
    fname << name1 << "/" << tmp2;
    int ret = stat(fname.getString(encoding),&sret);
    if (ret < 0){
      continue;
    }
    linkTarget=NULL;
    if ((sret.st_mode & S_IFMT) != S_IFREG){
	isFile=False;
    }else if( linklist && *linklist ){
      int len;
      len=strlen(finddata->d_name);
      if( len>=4 && !strcasecmp( finddata->d_name+len-4, ".LNK" ) ){
	w32shortcut *s=new w32shortcut( fname.getString(encoding) );
	if( s!=NULL && s->getStatus()!=0 ){
	  if( s->isDir() ){
	    isFile=False;
	  }else{
	    linkTarget=new WSCstring(s->getTarget(),_link_encoding);
	    linkTarget->replaceString("\\","/",0);
	  }
	  delete s;
	}
      }
    }
    if ( isFile ){
      WSCstring tmp;
#ifdef MacOS
      if (WSGIappLocaleSet()->getSystemLocaleEncoding() == WS_EN_UTF8){
extern char* macutf8_to_utf8(char*);
        char* t2 = macutf8_to_utf8(finddata->d_name);
        if (t2 != NULL){
          tmp.setString(t2,encoding);
          delete t2;
        }else{
          tmp.setString(finddata->d_name,encoding);
        }
      }else{
        tmp.setString(finddata->d_name,encoding);
      }
#else
      tmp.setString(finddata->d_name,encoding);
#endif
      list->add((void*)WSGFstrdup(tmp.getString()));
      if( linkTarget ){
	(*linklist)->setData(tmp.getString(),
	     (void*)linkTarget->getString((WS_EN_DEFAULT)));
      }
    }
  }
  closedir(dptr);
  WSGFsortCharList(list);
  return list;
#if 0
  char  buf[8192];
  sprintf(buf,"/bin/ls %s",name);

  FILE* pp;
  if (!(pp = popen(buf,"r"))){
    return NULL;
  }
  WSClistData* list = new WSClistData;

  while(fgets(buf,sizeof(buf),pp)){
    WSCstring str(buf);
    str.delString("\n",0);
    WSCstring str2(name);
    str2.addString("/");
    str2.addString(str);
    long iret = chk_file(str2.getString());
    if (iret  != WS_DIR && iret != WS_CANNOT){
      list->add((void*)WSGFstrdup(str.getString()));
    }
  }
  pclose(pp);
  return list;
#endif
}


WSCstring *WSDunixFS::readLink(char*linkname){
  WSCstring *p=new WSCstring(linkname);
  w32shortcut s(p->getString(WSGIappLocaleSet()
			     ->getSystemLocaleEncoding()));
  if( s.getStatus()!=0 ){
    char *target=s.getTarget();
    if( target ){
      p->setString(target, _link_encoding );
      p->replaceString("\\","/",0);
      return p;
    }
  }
  return NULL;
}

// w32shortcut definition start
static const unsigned char magic[]={
  'L',0,0,0,0x01,
  0x14,0x02,0x00,
  0,0,0,0,
  0xC0,0,0,0,
  0,0,0,0x46};

int
w32shortcut::getint( int off, int base, unsigned int len,
		     unsigned long *p, char *msg ){
    unsigned char buff[1024];
    unsigned long n=0;
    int r;
    int i;

    if( len>sizeof(buff) )return -1;

    r=fseek( fd, off, base );
    if( r<0 ){
      error_str=strerror(errno);
      return -1;
    }

    r=fread( buff, 1, len, fd );
    if( r<0 || (unsigned int)r<len ){
      error_str=strerror(errno);
      return -1;
    }
    for( i=len-1 ; i>=0 ; --i ){
	n=n*256+buff[i];
    }
    *p=n;
    return r;
}

#ifdef MSW
int
bcmp( const unsigned char *a, const unsigned char *b, int len ){
  int i;
  for( i=0 ; i<len && a[i]==b[i] ; i++ );
  return i>=len?0:a[i]>b[i]?1:-1;
}
#endif

w32shortcut::~w32shortcut(){
  delete name;
}

#define DEBUG 0

#define DBPRINTF if(DEBUG)fprintf

w32shortcut::w32shortcut( char *linkname ){
  unsigned char buff[1024];
  unsigned int i;

  status=0;
  error_str="";
  unicode=0;
  relPath=0;
  relPathLen=0;

  name=new char[strlen(linkname)+1];

  if( name==NULL ){
    error_str="insufficient memory";
    return;
  }

  strcpy(name,linkname);
#ifdef MSW
  fd=fopen(name,"rb");
#else
  fd=fopen(name,"r");
#endif
  if( fd==NULL ){
    error_str=strerror(errno);
    DBPRINTF(stderr,"%s (%s)\n",name,error_str);
    DBPRINTF(stderr,"%s %d:\n",__FILE__,__LINE__);
    return;
  }

  int l;
  int r;

  l=sizeof( magic );
  r=fread( buff, 1, l, fd );

  if( r<l || bcmp( buff, magic, l )!=0 ){
    if( r<0 ){
      error_str=strerror(errno);
      fclose( fd );
      DBPRINTF(stderr,"no magic(r=%d:%s)\n",r,error_str);
      return;
    }
    fclose( fd );
    error_str="no magic";
    return;
  }

  r=getint( 0, SEEK_CUR, 4, &flags1 );

  if( r<0 ){
    error_str=strerror(errno);
    fclose(fd);
    DBPRINTF(stderr,"no flag1(r=%d:%s)\n",r,error_str);
    return;
  }
  DBPRINTF(stderr,"%s:flags1=%08x:\n", name, flags1);

  r=getint( 0, SEEK_CUR, 4, &attr );

  if( r<0 ){
    error_str=strerror(errno);
    fclose(fd);
    DBPRINTF(stderr,"no attr(r=%d:%s)\n",r,error_str);
    return;
  }
  DBPRINTF(stderr,"%s:attr=%08x:\n", name, attr);

  r=fseek( fd, HEADER_END, SEEK_SET );
  if( r<0 ){
    error_str=strerror(errno);
    fclose(fd);
    DBPRINTF(stderr,"header skip error(r=%d:%s)\n",r,error_str);
    return;
  }

  if( flags1 & FLAGS1_SHELL_ITEM_ID_LIST ){
    unsigned long len;
    r=getint( 0, SEEK_CUR, 2, &len );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"shell item length error(r=%d:%s)\n",
	r,error_str);
      return;
    }
    r=fseek( fd, len, SEEK_CUR );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"shell item skip error(r=%d:%s)\n",
	r,error_str);
      return;
    }
  }

  if( flags1 & FLAGS1_FILE_OR_DIRECTORY ){
    long off0;
    unsigned long offhead, offheadoff;
    unsigned long offtail;
    unsigned long len2;

    off0=ftell( fd );

    if( off0<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"file_or_directory error(r=%d:%s)\n",
	r,error_str);
      return;
    }

    r=getint( 0, SEEK_CUR, 4, &len2 );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"file_or_directory length error(r=%d:%s)\n",
	r,error_str);
      return;
    }

    r=getint( off0+8, SEEK_SET, 4, &flags2 );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"file_or_directory flag2 error(r=%d:%s)\n",
	r,error_str);
      return;
    }

    if( flags2&(FLAGS2_LOCAL|FLAGS2_NETWORK)==0 ){
      fclose(fd);
      error_str="illegal location flag";
      DBPRINTF(stderr,"%s:illegal location flag %08x\n",
		name,flags2);
      return;
    }
    if( flags2&FLAGS2_LOCAL ){
      offheadoff=0x10;
    }else{
      offheadoff=0x14;
    }
    r=getint( off0+offheadoff, SEEK_SET, 4, &offhead );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,
	"file_or_directory head location error(r=%d:%s)\n",
	r,error_str);
      return;
    }
    r=getint( off0+0x18, SEEK_SET, 4, &offtail );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,
	"file_or_directory tail location error(r=%d:%s)\n",
	r,error_str);
      return;
    }
    if( !(flags2&FLAGS2_LOCAL) ){
      /* to skip header the part of network volume table */
      offhead+=0x14;
    }
    //(global-set-key "\C-Cf" 'word-search-forward)

    i=0;
    target[i]=0;
    r=fseek( fd, off0+offhead, SEEK_SET );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,
	"file_or_directory head part seek error(r=%d:%s)\n",
	r,error_str);
      return;
    }
    while( (r=getc(fd))!=EOF ){
      if( i<sizeof(target) ){
	target[i++]=r;
	if( r==0 )break;
      }else{
	fclose(fd);
	error_str="link target too long\n";
	DBPRINTF(stderr,"link target too long\n");
	return;
      }
    }

    if( !(flags2&FLAGS2_LOCAL) ){
      if( i>0 ){
	if( i<sizeof(target) ){
	  target[i-1]='\\';
	  target[i++]=0;
	}else{
	  fclose(fd);
	  error_str="link target too long\n";
	  return;
	}
      }
    }

    r=fseek( fd, off0+offtail, SEEK_SET );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,
	"file_or_directory tail part seek error(r=%d:%s)\n",
	r,error_str);
      return;
    }

    if( i>0 )i--;
    while( (r=getc(fd))!=EOF ){
      if( i<sizeof(target) ){
	target[i++]=r;
	if( r==0 )break;
      }else{
	fclose(fd);
	error_str="link target too long";
	DBPRINTF(stderr,"link target too long\n");
	return;
      }
    }
    DBPRINTF(stderr,"target='%s'\n",target);

    r=fseek( fd, off0+len2, SEEK_SET );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,
	"file_or_directory end seek error(r=%d:%s)\n",
	r,error_str);
      return;
    }
    fclose(fd);
    status=1;
  DBPRINTF(stderr,"w32shortcut ok(2)(target='%s')\n",target);
    return;
  }// FILE_OR_DIRECTORY
#if 1
  if( flags1&FLAGS1_DESC_STRING ){
    unsigned long off0, len;

    off0=ftell( fd );

    if( off0<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"description error(r=%d:%s)\n",
	r,error_str);
      return;
    }

    r=getint( 0, SEEK_CUR, 4, &len );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"description length error(r=%d:%s)\n",
	r,error_str);
      return;
    }

    if( flags1 & FLAGS1_UNICODE ){
      r=fseek( fd, len*2, SEEK_CUR );/* *2 for unicode */
    }else{
      r=fseek( fd, len, SEEK_CUR );/* for not unicode */
    }
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"description skip error(r=%d:%s)\n",
	r,error_str);
      return;
    }
  }
  if( flags1&FLAGS1_RELATIVE_PATH ){
    unsigned long off0, len;

    off0=ftell( fd );

    if( off0<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"relative path error(r=%d:%s)\n",
	r,error_str);
      return;
    }

    r=getint( 0, SEEK_CUR, 4, &len );
    if( r<0 ){
      error_str=strerror(errno);
      fclose(fd);
      DBPRINTF(stderr,"relative path length error(r=%d:%s)\n",
	r,error_str);
      return;
    }
    if( flags1 & FLAGS1_UNICODE ){
      len*=2;
    }
    if( len>sizeof(target) ){
      fclose(fd);
      error_str="link target relative path too long";
      return;
    }
    r=fread( target, 1, len, fd );
    if( r<0 || (unsigned int)r<len ){
      if( r<0 ){
	error_str=strerror(errno);
      }else{
	error_str="relative path read error";
      }
      fclose(fd);
      DBPRINTF(stderr,"relative path skip error(r=%d:%s)\n",
	r,error_str);
      return;
    }
    relPath=1;
    relPathLen=len;
    if( flags1 & FLAGS1_UNICODE ){
      unicode=1;
    }
    fclose(fd);
    status=1;
  DBPRINTF(stderr,"w32shortcut ok(3)(target='%s')\n",target);
    return;
  }
  /* WORKING_DIRECTORY */
  /* COMMAND_LINE_ARGUMENTS */
  /* CUSTOM_ICON */
  fclose(fd);
  status=0;
#endif
  DBPRINTF(stderr,"w32shortcut ok(1)(target='%s')\n",target);
}
// w32shortcut definition end
