//
// Copyright (C) 1999-2008 WideStudio/MWT Project Team
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#include <x11/WSDxappDev.h>
#include <WScom.h>
#include <WStcpcom.h>
#include <x11/WSxcom.h>
#include <x11/WSDunixSemaphore.h>
#include <stdio.h>
#include <fcntl.h>
#include <WSClocaleSet.h>
#include <pthread.h>
#ifndef NO_IPC_SEM
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#endif

WSMFclassInit(WSDunixSemaphore,WSDsemaphore);

WSDsemaphore* _unix_sem_chandler(){
  return new WSDunixSemaphore;
}

void WSGFinitializeUnixSemaphore(){
    WSDsemaphore::setCreateInstanceHandler((void*)_unix_sem_chandler);
}
#ifndef NO_GLOBAL_CONSTRUCTORS
class _unix_sem_init {
  public: _unix_sem_init(){
    WSGFinitializeUnixSemaphore();
  };
};
_unix_sem_init  _unix_sem_init_execute;
#endif

WSDunixSemaphore::WSDunixSemaphore(){
  _semid = -1;
}

WSDunixSemaphore::~WSDunixSemaphore(){
}

#ifndef NO_IPC_SEM
#ifndef __FreeBSD__
union semun {
  int val;
  struct semid_ds *buf;
  unsigned short *array;
  struct seminfo *__buf;
};
#endif
#endif
long WSDunixSemaphore::initialize(){
  if (_private != False){
    pthread_sem_init_np(&_sem);
  }else{
#ifndef NO_IPC_SEM
    if (_semid != -1){
      destroy();
    }
    long val = 1000;
    WSCstring key(_key);
    if(_key.isExist(":") != -1){
      WSCstring valstr = _key.getWord(1,":");
      val = WSGFatol(valstr.getString());
      key = _key.getWord(0,":");
    }
    _sem_key = ftok(key.getString(),val);
    if (_sem_key == -1){
      return WS_ERR;
    }
    if (_owner != False){
      if ((_semid = semget(_sem_key,1,0666|IPC_CREAT)) == -1){
        return WS_ERR;
      }
      union semun arg;
      arg.val = 1;
      if (semctl(_semid,0,SETVAL,arg) == -1){
        return WS_ERR;
      }
    }else{
      if ((_semid = semget(_sem_key,1,0)) == -1){
        return WS_ERR;
      }
    }
#else
    return WS_ERR;
#endif
  }
  return WS_NO_ERR;
}
long WSDunixSemaphore::lock(){
  if (_private != False){
    int ret = pthread_sem_lock_np(&_sem);
    if (ret == 0){
      return WS_NO_ERR;
    }
    return WS_ERR;
  }else{
#ifndef NO_IPC_SEM
    struct sembuf semb = {0,-1,0};
    if (semop(_semid,&semb,1) == -1){
      return WS_ERR;
    }
    return WS_NO_ERR;
#else
    return WS_ERR;
#endif
  }
  return WS_NO_ERR;
}

long WSDunixSemaphore::unlock(){
  if (_private != False){
    int ret = pthread_sem_unlock_np(&_sem);
    if (ret == 0){
      return WS_NO_ERR;
    }
    return WS_ERR;
  }else{
#ifndef NO_IPC_SEM
    struct sembuf semb = {0,1,0};
    if (semop(_semid,&semb,1) == -1){
      return WS_ERR;
    }
    return WS_NO_ERR;
#else
    return WS_ERR;
#endif
  }
}

long WSDunixSemaphore::destroy(){
  if (_private != False){
    int ret = pthread_sem_destroy_np(&_sem);
    if (ret == 0){
      return WS_NO_ERR;
    }
    return WS_ERR;
  }else{
#ifndef NO_IPC_SEM
    if (_semid == -1){
      return WS_ERR;
    }
    if (_owner != False){
      union semun arg;
      arg.val = 1;
      if (semctl(_semid,0,IPC_RMID,arg) == -1){
        _semid = -1;
        return WS_ERR;
      }
      _semid = -1;
      return WS_NO_ERR;
    }
#endif
    return WS_ERR;
  }
}
